#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Dict

import firefly_di as di
from firefly.domain.factory.factory import T
from firefly.domain.service.core.agent import Agent

from .factory import Factory
from ..error import ConfigurationError


class AgentFactory(Factory[Agent]):
    def __init__(self, agents: Dict[str, Agent]):
        self._agents = agents

    def __call__(self, provider: str) -> T:
        if provider not in self._agents:
            raise ConfigurationError(f'No agent registered for provider "{provider}"')
        return self._agents[provider]

    def register(self, provider: str, agent: Agent):
        self._agents[provider] = agent



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import TypeVar, Type

import firefly.domain as ffd

from .repository import Repository
from ..entity.aggregate_root import AggregateRoot
from ..service.logging.logger import LoggerAware

AR = TypeVar('AR', bound=AggregateRoot)


class Registry(LoggerAware):
    def __init__(self):
        self._cache = {}
        self._factories = {}
        self._default_factory = {}

    def __call__(self, entity) -> Repository:
        if not issubclass(entity, ffd.AggregateRoot):
            raise ffd.LogicError('Repositories can only be generated for aggregate roots')

        if entity not in self._cache:
            for k, v in self._factories.items():
                if issubclass(entity, k):
                    self._cache[entity] = v(entity)
                    break

            context = entity.get_class_context()
            if context in self._default_factory and self._default_factory[context] is not None:
                self._cache[entity] = self._default_factory[context](entity)

            if entity not in self._cache:
                raise ffd.FrameworkError(
                    f'No registry found for entity {entity}. Have you configured a persistence mechanism or extension?'
                )

        return self._cache[entity]

    def register_factory(self, type_: Type[AR], factory: ffd.RepositoryFactory):
        self._factories[type_] = factory

    def set_default_factory(self, context: str, factory: ffd.RepositoryFactory):
        self._default_factory[context] = factory

    def clear_cache(self):
        self._cache = {}

    def get_repositories(self):
        ret = []
        for k, v in self._cache.items():
            ret.append(v)
        self.debug(ret)
        return ret



#!/usr/bin/env python
#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

"""These functions are executed via gyp-flock-tool when using the Makefile
generator.  Used on systems that don't have a built-in flock."""

import fcntl
import os
import struct
import subprocess
import sys


def main(args):
  executor = FlockTool()
  executor.Dispatch(args)


class FlockTool(object):
  """This class emulates the 'flock' command."""
  def Dispatch(self, args):
    """Dispatches a string command to a method."""
    if len(args) < 1:
      raise Exception("Not enough arguments")

    method = "Exec%s" % self._CommandifyName(args[0])
    getattr(self, method)(*args[1:])

  def _CommandifyName(self, name_string):
    """Transforms a tool name like copy-info-plist to CopyInfoPlist"""
    return name_string.title().replace('-', '')

  def ExecFlock(self, lockfile, *cmd_list):
    """Emulates the most basic behavior of Linux's flock(1)."""
    # Rely on exception handling to report errors.
    # Note that the stock python on SunOS has a bug
    # where fcntl.flock(fd, LOCK_EX) always fails
    # with EBADF, that's why we use this F_SETLK
    # hack instead.
    fd = os.open(lockfile, os.O_WRONLY|os.O_NOCTTY|os.O_CREAT, 0666)
    if sys.platform.startswith('aix'):
      # Python on AIX is compiled with LARGEFILE support, which changes the
      # struct size.
      op = struct.pack('hhIllqq', fcntl.F_WRLCK, 0, 0, 0, 0, 0, 0)
    else:
      op = struct.pack('hhllhhl', fcntl.F_WRLCK, 0, 0, 0, 0, 0, 0)
    fcntl.fcntl(fd, fcntl.F_SETLK, op)
    return subprocess.call(cmd_list)


if __name__ == '__main__':
  sys.exit(main(sys.argv[1:]))



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import firefly as ff

# __pragma__('skip')
from abc import ABC
# __pragma__('noskip')
# __pragma__ ('ecom')

"""?
from firefly.presentation.web.polyfills import ABC
?"""
# __pragma__ ('noecom')

from firefly.domain.entity.entity import Entity, hidden
from firefly.domain.meta.meta_aware import MetaAware


class AggregateRoot(Entity, MetaAware, ABC):
    # __pragma__('skip')
    _create_on: ff.TypeOfEvent = hidden()
    _delete_on: ff.TypeOfEvent = hidden()
    _update_on: ff.TypeOfEvent = hidden()

    @classmethod
    def get_create_on(cls):
        return cls._create_on

    @classmethod
    def get_delete_on(cls):
        return cls._delete_on

    @classmethod
    def get_update_on(cls):
        return cls._update_on
    # __pragma__('noskip')



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import fields
from datetime import datetime
from typing import Type

import firefly.domain as ffd
import inflection


class RdbStorageInterface(ffd.LoggerAware, ABC):
    _serializer: ffd.Serializer = None
    _cache: dict = {}

    def __init__(self):
        self._tables_checked = []
        self._cache = {
            'sql': {
                'insert': {},
                'update': {},
            },
            'indexes': {},
            'parts': {
                'columns': {},
                'values': {},
                'update': {},
                'select': {},
            }
        }

    def disconnect(self):
        self._disconnect()

    @abstractmethod
    def _disconnect(self):
        pass

    def add(self, entity: ffd.Entity):
        self._check_prerequisites(entity.__class__)
        self._add(entity)

    @abstractmethod
    def _add(self, entity: ffd.Entity):
        pass

    def all(self, entity_type: Type[ffd.Entity], criteria: ffd.BinaryOp = None, limit: int = None):
        self._check_prerequisites(entity_type)
        return self._all(entity_type, criteria, limit)

    @abstractmethod
    def _all(self, entity_type: Type[ffd.Entity], criteria: ffd.BinaryOp = None, limit: int = None):
        pass

    def find(self, uuid: str, entity_type: Type[ffd.Entity]):
        self._check_prerequisites(entity_type)
        return self._find(uuid, entity_type)

    @abstractmethod
    def _find(self, uuid: str, entity_type: Type[ffd.Entity]):
        pass

    def remove(self, entity: ffd.Entity):
        self._check_prerequisites(entity.__class__)
        self._remove(entity)

    @abstractmethod
    def _remove(self, entity: ffd.Entity):
        pass

    def update(self, entity: ffd.Entity):
        self._check_prerequisites(entity.__class__)
        self._update(entity)

    @abstractmethod
    def _update(self, entity: ffd.Entity):
        pass

    @abstractmethod
    def _ensure_connected(self):
        pass

    @staticmethod
    def _fqtn(entity: Type[ffd.Entity]):
        return inflection.tableize(entity.get_fqn())

    def _check_prerequisites(self, entity: Type[ffd.Entity]):
        self._ensure_connected()

    def get_indexes(self, entity: Type[ffd.Entity]):
        if entity not in self._cache['indexes']:
            self._cache['indexes'][entity] = []
            for field_ in fields(entity):
                if 'index' in field_.metadata and field_.metadata['index'] is True:
                    self._cache['indexes'][entity].append(field_)

        return self._cache['indexes'][entity]

    def _generate_insert(self, entity: ffd.Entity):
        t = entity.__class__
        sql = f"insert into {self._fqtn(t)} ({self._generate_column_list(t)}) values ({self._generate_value_list(t)})"
        print(sql)
        print(self._generate_parameters(entity))
        return sql, self._generate_parameters(entity)

    def _generate_update(self, entity: ffd.Entity):
        t = entity.__class__
        sql = f"update {self._fqtn(t)} set {self._generate_update_list(t)} where id = :id"
        return sql, self._generate_parameters(entity)

    @abstractmethod
    def _generate_update_list(self, entity: Type[ffd.Entity]):
        pass

    @abstractmethod
    def _generate_column_list(self, entity: Type[ffd.Entity]):
        pass

    @abstractmethod
    def _generate_select_list(self, entity: Type[ffd.Entity]):
        pass

    @abstractmethod
    def _generate_value_list(self, entity: Type[ffd.Entity]):
        pass

    @abstractmethod
    def _generate_parameters(self, entity: ffd.Entity, part: str = None):
        pass

    @abstractmethod
    def _build_entity(self, entity: Type[ffd.Entity], data, raw: bool = False):
        pass

    @staticmethod
    def _generate_where_clause(criteria: ffd.BinaryOp):
        if criteria is None:
            return '', {}
        return criteria.to_sql()

    @staticmethod
    def _generate_index(name: str):
        return ''

    @staticmethod
    def _db_type(field_):
        if field_.type == 'float' or field_.type is float:
            return 'float'
        if field_.type == 'integer' or field_.type is int:
            return 'integer'
        if field_.type == 'datetime' or field_.type is datetime:
            return 'datetime'
        length = field_.metadata['length'] if 'length' in field_.metadata else 256
        return f'varchar({length})'

    @abstractmethod
    def _generate_create_table(self, entity: Type[ffd.Entity]):
        pass

    @staticmethod
    def _datetime_declaration(name: str):
        return f"`{name}` datetime"

    @staticmethod
    def _generate_extra(columns: list, indexes: list):
        return f", {','.join(columns)}"

    def execute_ddl(self, entity: Type[ffd.Entity]):
        self._execute_ddl(entity)

    @abstractmethod
    def _execute_ddl(self, entity: Type[ffd.Entity]):
        pass

    @abstractmethod
    def raw(self, entity: Type[ffd.Entity], criteria: ffd.BinaryOp = None, limit: int = None):
        pass



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Any

import firefly.domain as ffd

from .configuration_annotation import ConfigurationAnnotation


class Authenticator(ConfigurationAnnotation):
    def __call__(self, *args, **kwargs):
        def authenticator_wrapper(cls: ffd.MetaAware):
            try:
                cls.add_annotation(self)
            except AttributeError:
                raise ffd.FrameworkError('@authenticator used on invalid target')
            return cls

        return authenticator_wrapper

    def configure(self, cls: Any, container):
        if not issubclass(cls, ffd.Handler):
            raise ffd.ConfigurationError('Authenticators must be a subclass of Handler')
        container.authenticator.add(container.build(cls))


authenticator = Authenticator()



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from dataclasses import asdict
from typing import TypeVar, Generic, Type

import firefly.domain as ffd

from .crud_operation import CrudOperation
from ..core.application_service import ApplicationService
from ..messaging.system_bus import SystemBusAware
from ...value_object.generic_base import GenericBase

T = TypeVar('T')


class CreateEntity(Generic[T], ApplicationService, GenericBase, CrudOperation, SystemBusAware):
    _registry: ffd.Registry = None
    _context_map: ffd.ContextMap = None

    def __call__(self, **kwargs) -> ffd.Entity:
        type_ = self._type()
        method = self._find_factory_method(type_)
        if method is not None:
            entity = method(type_, **kwargs)
        else:
            entity = type_(**ffd.build_argument_list(kwargs, type_))
        self._registry(type_).append(entity)
        self.dispatch(self._build_event(type_, 'create', asdict(entity), kwargs['_context']))

        return entity

    @staticmethod
    def _find_factory_method(type_: Type[ffd.Entity]):
        for k, v in type_.__dict__.items():
            if k == 'create' and isinstance(v, classmethod):
                return v.__func__
        return None



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import importlib
import sys
from dataclasses import dataclass
from time import sleep

import firefly as ff


def retry(cb, valid_cb=None, wait: int = 1, backoff: bool = True, retries: int = 5, catch=Exception, should_retry=None):
    while retries > 0:
        try:
            ret = cb()
            if valid_cb is not None and not valid_cb(ret):
                raise RuntimeError('valid_cb is false-y for return value')
            return ret
        except catch as e:
            if should_retry is not None and not should_retry(e):
                raise e
            retries -= 1
            if retries <= 0:
                raise e
            sleep(wait)
            if backoff:
                wait *= 2


def load_class(fqn: str):
    pieces = fqn.split('.')
    module_name = '.'.join(pieces[0:-1])
    class_ = pieces[-1]

    if module_name not in sys.modules:
        try:
            module = importlib.import_module(module_name)
            return getattr(module, class_)
        except (ImportError, AttributeError):
            return None

    try:
        return getattr(sys.modules[module_name], class_)
    except AttributeError:
        return None


def generate_dc(base: type, _cls, **kwargs):
    if 'eq' not in kwargs:
        kwargs['eq'] = False
    if 'repr' not in kwargs:
        kwargs['repr'] = False

    def wrapper(cls):
        dc = dataclass(cls, **kwargs)
        if issubclass(dc, base):
            return dc

        class Wrapper(dc, base):
            pass
        Wrapper.__name__ = cls.__name__

        return dataclass(Wrapper, **kwargs)

    if _cls is None:
        return wrapper

    return wrapper(_cls)


class ValueMeta(type):
    def __new__(mcs, name, bases, dct, **kwargs):
        ret = type.__new__(mcs, name, bases, dct)
        return dataclass(ret, frozen=True)


def has_meta(cls):
    return hasattr(cls, '__FIREFLY__')


def check_for_meta(cls):
    if not has_meta(cls):
        cls.__FIREFLY__ = type.__new__(type, f'{cls.__name__}Meta', (ff.MetaAware,), {})


def add_endpoint(cls, endpoint: ff.Endpoint):
    check_for_meta(cls)
    cls.__FIREFLY__.add_endpoint(endpoint)


def add_event(cls, event: ff.TypeOfEvent):
    check_for_meta(cls)
    cls.__FIREFLY__.add_event(event)


def set_command(cls, command: ff.TypeOfCommand):
    check_for_meta(cls)
    cls.__FIREFLY__.set_command(command)


def set_query(cls, query: ff.TypeOfQuery):
    check_for_meta(cls)
    cls.__FIREFLY__.set_query(query)


def get_endpoints(cls):
    check_for_meta(cls)
    return cls.__FIREFLY__.get_endpoints()


def get_events(cls):
    check_for_meta(cls)
    return cls.__FIREFLY__.get_events()


def get_command(cls):
    check_for_meta(cls)
    return cls.__FIREFLY__.get_command()


def get_query(cls):
    check_for_meta(cls)
    return cls.__FIREFLY__.get_query()


def has_endpoints(cls):
    check_for_meta(cls)
    return cls.__FIREFLY__.has_endpoints()


def is_event_listener(cls):
    check_for_meta(cls)
    return cls.__FIREFLY__.is_event_listener()


def is_command_handler(cls):
    check_for_meta(cls)
    return cls.__FIREFLY__.is_command_handler()


def is_query_handler(cls):
    check_for_meta(cls)
    return cls.__FIREFLY__.is_query_handler()


def merge(a, b, path=None):
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass
            else:
                a[key] = b[key]
                # raise Exception('Conflict at %s' % '.'.join(path + [str(key)]))
        else:
            a[key] = b[key]
    return a



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

# __pragma__('skip')
from __future__ import annotations

import typing
from dataclasses import asdict
from typing import List

from firefly.domain.meta.build_argument_list import build_argument_list

from .event_buffer import EventBuffer
from .generic_base import GenericBase
from .parameter import Parameter

from firefly.domain.meta.entity_meta import EntityMeta


class ValueObject(metaclass=EntityMeta):
    _logger = None

    def __init__(self, **kwargs):
        pass

    def to_dict(self):
        # noinspection PyDataclass
        return asdict(self)

    def debug(self, *args, **kwargs):
        return self._logger.debug(*args, **kwargs)

    def info(self, *args, **kwargs):
        return self._logger.info(*args, **kwargs)

    def warning(self, *args, **kwargs):
        return self._logger.warning(*args, **kwargs)

    def error(self, *args, **kwargs):
        return self._logger.error(*args, **kwargs)

    @classmethod
    def from_dict(cls, data: dict, map_: dict = None):
        d = None
        if map_ is not None:
            d = data.copy()
            for source, target in map_.items():
                d[target] = d[source]
        return cls(**build_argument_list(d or data, cls))
    # __pragma__('noskip')



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

# __pragma__('js', '{}', "export const inflection = require('inflection');")
# __pragma__('skip')


class Inflection:
    def indexOf(self, arr: list, item, from_index: int = None, compare_func=None):
        pass

    def pluralize(self, string: str, plural: str = None):
        pass

    def singularize(self, string: str, singular: str = None):
        pass

    def inflect(self, string: str, count: int, singular: str = None, plural: str = None):
        pass

    def camelize(self, string: str, low_first_letter: bool = False):
        pass

    def underscore(self, string: str, all_upper_case: bool = False):
        pass

    def humanize(self, string: str, low_first_letter: bool = False):
        pass

    def capitalize(self, string: str):
        pass

    def dasherize(self, string: str):
        pass

    def titleize(self, string: str):
        pass

    def demodulize(self, string: str):
        pass

    def tableize(self, string: str):
        pass

    def classify(self, string: str):
        pass

    def foreign_key(self, string: str, drop_id_ubar: bool = False):
        pass

    def ordinalize(self, string: str):
        pass

    def transform(self, string: str, arr: list):
        pass


inflection = Inflection()



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import typing
from datetime import datetime, date

import inflection

from firefly.domain.meta.entity_meta import EntityMeta
from firefly.domain.value_object import ValueObject
from firefly.domain.meta.context_aware import ContextAware
from firefly.domain.meta.build_argument_list import build_argument_list

# __pragma__('skip')
import uuid

from dataclasses import is_dataclass, fields, field, MISSING, asdict
from typing import List, Callable
from abc import ABC
# __pragma__('noskip')
# __pragma__ ('ecom')
"""?
from firefly.presentation.web.polyfills import is_dataclass, fields, field, MISSING, asdict, List, Callable, uuid
?"""
# __pragma__ ('noecom')
# __pragma__('kwargs')


# noinspection PyDataclass
class Entity(ContextAware, ValueObject):
    _logger = None

    def __init__(self, **kwargs):
        pass

    def __post_init__(self):
        if is_dataclass(self):
            missing = []
            for field_ in fields(self):
                if 'required' in field_.metadata and isinstance(getattr(self, field_.name), Empty):
                    missing.append(field_.name)
            if len(missing) > 0:
                raise TypeError(f'__init__ missing {len(missing)} required argument(s): {", ".join(missing)}')

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        return self.id_value() == other.id_value()

    def id_value(self):
        if not is_dataclass(self):
            raise TypeError('Entity::id_value() called on a non-dataclass entity')

        for field_ in fields(self):
            if 'id' in field_.metadata:
                return getattr(self, field_.name)

    def load_dict(self, data: dict):
        data = self._process_data(self.__class__, data)
        t = typing.get_type_hints(self.__class__)
        for name, type_ in t.items():
            if name in data:
                setattr(self, name, data[name])

    @staticmethod
    def _process_data(cls, data: dict):
        t = typing.get_type_hints(cls)
        for name, type_ in t.items():
            if name.startswith('_'):
                continue

            if isinstance(type_, type(List)) and len(type_.__args__) == 1:
                new_list = []
                for item in data[name]:
                    nested_data = Entity._process_data(type_.__args__[0], item)
                    new_list.append(type_.__args__[0](**build_argument_list(nested_data, type_.__args__[0])))
                data[name] = new_list
            else:
                try:
                    if issubclass(type_, Entity):
                        data[name] = type_(**build_argument_list(data[name], type_))
                except TypeError:
                    pass

        return data

    @classmethod
    def id_name(cls):
        if not is_dataclass(cls):
            raise TypeError('Entity::id_name() called on a non-dataclass entity')

        for field_ in fields(cls):
            if 'id' in field_.metadata:
                return field_.name

    @classmethod
    def match_id_from_argument_list(cls, args: dict):
        snake = f'{inflection.underscore(cls.__name__)}_id'
        if snake in args:
            return {snake: args[snake]}

        id_name = cls.id_name()
        if id_name in args:
            return {id_name: args[id_name]}

    def debug(self, *args, **kwargs):
        return self._logger.debug(*args, **kwargs)

    def info(self, *args, **kwargs):
        return self._logger.info(*args, **kwargs)

    def warning(self, *args, **kwargs):
        return self._logger.warning(*args, **kwargs)

    def error(self, *args, **kwargs):
        return self._logger.error(*args, **kwargs)


class Empty:
    pass


def id_(is_uuid: bool = True, **kwargs):
    metadata = {'id': True, 'required': True, 'type': str}
    if is_uuid:
        metadata['length'] = 36
    metadata.update(kwargs)
    return field(default_factory=lambda: str(uuid.uuid4()), metadata=metadata) if is_uuid else \
        required(**metadata)


def list_(**kwargs):
    kwargs['type'] = list
    return field(default_factory=lambda: [], metadata=kwargs)


def dict_(**kwargs):
    kwargs['type'] = dict
    return field(default_factory=lambda: {}, metadata=kwargs)


def now(**kwargs):
    kwargs['type'] = datetime
    return field(default_factory=lambda: datetime.now(), metadata=kwargs)


def today(**kwargs):
    kwargs['type'] = date
    return field(default_factory=lambda: date.today(), metadata=kwargs)


def required(type_: type = None, **kwargs):
    if type_ is not None:
        kwargs['type'] = type_
    kwargs['required'] = True
    return field(default_factory=lambda: Empty(), metadata=kwargs)


def optional(type_: type = None, default=MISSING, **kwargs):
    if type_ is not None:
        kwargs['type'] = type_
    kwargs['required'] = False
    if not isinstance(default, MISSING.__class__):
        return field(default_factory=lambda: default, metadata=kwargs)
    return field(default=None, metadata=kwargs)


def hidden(**kwargs):
    kwargs['hidden'] = True
    return field(default=None, init=False, repr=False, compare=False, metadata=kwargs)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import importlib
import inspect
import typing
from typing import List, Type

import firefly.domain as ffd
import firefly_di as di
import inflection


class LoadApplicationLayer(ffd.ApplicationService):
    _context_map: ffd.ContextMap = None
    _event_resolver: ffd.EventResolvingMiddleware = None
    _command_resolver: ffd.CommandResolvingMiddleware = None
    _query_resolver: ffd.QueryResolvingMiddleware = None
    _agent_factory: ffd.AgentFactory = None
    _message_factory: ffd.MessageFactory = None
    _rest_router: ffd.RestRouter = None
    _system_bus: ffd.SystemBus = None
    _deferred: List[tuple] = []

    def __call__(self, **kwargs):
        self._deferred = []
        for context in self._context_map.contexts:
            for cls in self._load_module(context):
                if cls.has_annotations():
                    for annotation in cls.get_annotations():
                        annotation.configure(cls, context.container)
                if issubclass(cls, ffd.ApplicationService):
                    self._register_service(cls, context)
                    self._add_endpoints(cls, context)
                elif issubclass(cls, ffd.Middleware):
                    self._add_middleware(cls, context)
            for entity in context.entities:
                if issubclass(entity, ffd.AggregateRoot):
                    self._add_endpoints(entity, context)

        for deferred in self._deferred:
            if deferred[0] == 'event':
                self._system_bus.add_event_listener(deferred[1], **deferred[2])
            elif deferred[0] == 'command':
                self._system_bus.add_command_handler(deferred[1], **deferred[2])
            elif deferred[0] == 'query':
                self._system_bus.add_query_handler(deferred[1], **deferred[2])

        self.dispatch(ffd.ApplicationLayerLoaded())

    def _add_middleware(self, cls: Type[ffd.Middleware], context: ffd.Context):
        if not cls.is_middleware():
            return
        config = cls.get_middleware_config()
        params = {}
        if config['index'] is not None:
            params['index'] = config['index']
        elif config['cb'] is not None:
            params['cb'] = config['cb']
        if config['replace'] is not None:
            params['replace'] = config['replace']

        built = context.container.build(cls)
        if config['buses'] is None or 'event' in config['buses']:
            if self._system_bus.add_event_listener(built, **params) is False:
                self._deferred.append(('event', built, params))
        if config['buses'] is None or 'command' in config['buses']:
            if self._system_bus.add_command_handler(built, **params) is False:
                self._deferred.append(('command', built, params))
        if config['buses'] is None or 'query' in config['buses']:
            if self._system_bus.add_query_handler(built, **params) is False:
                self._deferred.append(('query', built, params))

    def _register_service(self, cls: Type[ffd.ApplicationService], context: ffd.Context):
        if cls.is_agent():
            self._agent_factory.register(cls.get_agent(), context.container.build(cls))

        if not cls.is_handler():
            return

        if cls.is_event_listener() and issubclass(cls, ffd.ApplicationService):
            for event in cls.get_events():
                self._event_resolver.add_event_listener(cls, event)
                if cls not in context.event_listeners:
                    context.event_listeners[cls] = []
                context.event_listeners[cls].append(event)

        if cls.is_command_handler() and issubclass(cls, ffd.ApplicationService):
            if cls.get_command() is None:
                cls.set_command(self._generate_message_name(cls))
            cmd = cls.get_command()
            if isinstance(cmd, str):
                cmd = self._message_factory.command_class(cmd, typing.get_type_hints(cls.__call__))
            self._command_resolver.add_command_handler(cls, cmd)
            context.command_handlers[cls] = cmd

        if cls.is_query_handler() and issubclass(cls, ffd.ApplicationService):
            if cls.get_query() is None:
                cls.set_query(self._generate_message_name(cls))
            query = cls.get_query()
            if isinstance(query, str):
                query = self._message_factory.query_class(query, typing.get_type_hints(cls.__call__))
            self._query_resolver.add_query_handler(cls, query)
            context.query_handlers[cls] = query

    def _add_endpoints(self, cls: Type[ffd.MetaAware], context: ffd.Context):
        if not cls.has_endpoints():
            return

        for endpoint in cls.get_endpoints():
            if endpoint.message is None:
                if cls in context.command_handlers:
                    endpoint.message = context.command_handlers[cls]
                elif cls in context.query_handlers:
                    endpoint.message = context.query_handlers[cls]
                elif isinstance(endpoint, ffd.HttpEndpoint):
                    if endpoint.method.lower() == 'get':
                        cls.set_query(self._generate_message(cls, 'query'))
                        endpoint.message = cls.get_query()
                    else:
                        cls.set_command(self._generate_message(cls, 'command'))
                        endpoint.message = cls.get_command()
                    self._register_service(cls, context)
                elif isinstance(endpoint, ffd.CliEndpoint):
                    cls.set_command(self._generate_message(cls, 'command'))
                    endpoint.message = cls.get_command()
                    self._register_service(cls, context)

            context.endpoints.append(endpoint)
            if isinstance(endpoint, ffd.HttpEndpoint):
                route_prefix = f'/{inflection.dasherize(context.name)}'
                route = endpoint.route
                if not route.startswith('/'):
                    route = f'/{route}'
                if not route.startswith(route_prefix):
                    route = f'{route_prefix}{route}'
                self._rest_router.register(route, endpoint)

    @staticmethod
    def _load_module(context: ffd.Context) -> List[Type[ffd.ApplicationService]]:
        module_name = context.config.get('application_module', '{}.application')
        try:
            module = importlib.import_module(module_name.format(context.name))
        except (ModuleNotFoundError, KeyError):
            return []

        ret = []
        for k, v in module.__dict__.items():
            if not inspect.isclass(v):
                continue
            if issubclass(v, (ffd.ApplicationService, ffd.Middleware, ffd.MetaAware)):
                ret.append(v)

        return ret

    @staticmethod
    def _generate_message_name(cls):
        return f'{cls.get_class_context()}.{inflection.camelize(cls.__name__)}'

    def _generate_message(self, cls, type_: str):
        return getattr(self._message_factory, f'{type_}_class')(
            self._generate_message_name(cls),
            fields_=typing.get_type_hints(cls.__call__)
        )



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

import argparse
import os

import inflection
from firefly.domain.meta.build_argument_list import build_argument_list
from firefly.domain.entity.messaging.command import Command
from firefly.domain.entity.messaging.query import Query
from firefly.domain.entity.core.cli_app import CliApp
from firefly.domain.entity.core.cli_argument import CliArgument
from firefly.domain.service.command_line.cli_app_executor import CliAppExecutor
from firefly.domain.service.messaging.message_factory import MessageFactory
from firefly.domain.service.messaging.system_bus import SystemBusAware


class ArgparseExecutor(CliAppExecutor, SystemBusAware):
    _message_factory: MessageFactory = None

    def __init__(self):
        self._app = None
        self._parser = None
        self._tree = {}
        self._message_cache = {}

    def run(self, app: CliApp):
        self._app = app
        params = {}
        if app.description is not None:
            params['description'] = app.description
        self._parser = argparse.ArgumentParser(**params)
        self._initialize()

        args = self._parser.parse_args()

        if args.verbose is True:
            self._logger.set_level_to_info()
        if args.debug is True:
            self._logger.set_level_to_debug()
        os.environ['ENV'] = args.env or 'local'

        try:
            target = args.target
        except AttributeError:
            target = None
        if target is None:
            self._parser.print_help()
            return

        message = self._message_cache[target]
        instance = message(**build_argument_list(vars(args), message))
        if issubclass(message, Command):
            self.invoke(instance)
        elif issubclass(message, Query):
            self.request(instance)

    def _initialize(self):
        self._add_verbosity_arguments(self._parser)
        self._build_node_tree()
        if 'children' not in self._tree:
            return self._configure_single_command_app()
        self._configure_argparse()

    def _build_node_tree(self):
        for endpoint in self._app.endpoints:
            parts = endpoint.command.split(' ')
            if parts[0] == self._app.name:
                parts.pop(0)
            node = self._tree
            if len(parts) > 1:
                for part in parts[:-1]:
                    if 'children' not in node:
                        node['children'] = {}
                    if part not in node['children']:
                        node['children'][part] = {}
                    node = node['children'][part]
            if 'children' not in node:
                node['children'] = {}
            if parts[-1] not in node['children']:
                node['children'][parts[-1]] = {}
            node['children'][parts[-1]]['endpoint'] = endpoint

    def _configure_argparse(self, node: dict = None, parser: argparse.ArgumentParser = None):
        if parser is None:
            parser = self._parser
        if node is None:
            node = self._tree

        if 'endpoint' in node:
            if len(node['endpoint'].arguments):
                for arg in node['endpoint'].arguments:
                    self._add_argument(parser, arg)

        if 'children' in node:
            sp = parser.add_subparsers(help='Commands', dest='target')
            for k, v in node['children'].items():
                p = sp.add_parser(k, help=v['endpoint'].help if 'endpoint' in v else None)
                if 'endpoint' in v:
                    self._message_cache[k] = v['endpoint'].message
                self._configure_argparse(v, p)

    @staticmethod
    def _add_argument(parser, arg: CliArgument):
        params = [f'--{inflection.dasherize(arg.name)}'] + [a for a in arg.alias or []]
        kwargs = {
            'default': arg.default,
            'required': arg.required,
            'help': arg.help,
        }
        if arg.help is None and arg.default is not None:
            kwargs['help'] = f'default: {arg.default}'
        if arg.type == bool:
            kwargs['action'] = 'store_true' if not arg.default else 'store_false'
        else:
            kwargs['type'] = arg.type
        parser.add_argument(*params, **kwargs)

    def _configure_single_command_app(self):
        if len(self._app.endpoints):
            for arg in self._app.endpoints[0].arguments:
                self._add_argument(self._parser, arg)

    @staticmethod
    def _add_verbosity_arguments(parser):
        try:
            parser.add_argument('--verbose', '-v', action='store_true', default=False, help='Enable verbose output')
            parser.add_argument('--debug', '-d', action='store_true', default=False, help='Enable debug output')
            parser.add_argument('--env', '-e', action='store', default='local', help='Environment name')
        except argparse.ArgumentError:
            pass



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from firefly.infrastructure import set_env


@set_env
def main():
    from firefly.application import Container
    container = Container()
    container.kernel.boot()
    container.cli_executor.run(
        container.context_map.get_cli_app('firefly')
    )


# class CliOutput(ffd.Middleware):
#     def __call__(self, message: ffd.Message, next_: Callable, **kwargs) -> Optional[dict]:
#         print('CliOutput called...')
#         response = next_(message)
#
#         if isinstance(response, dict):
#             for title, data in response.items():
#                 response[title].insert(0, ('Name', 'Type'))
#                 print(SingleTable(data, title).table)
#         elif isinstance(response, list):
#             response.insert(0, ('Name', 'Type'))
#             print(SingleTable(response).table)
#
#         return response
#
#
# @ffd.cli(
#     app_name='firefly',
#     description='Firefly command line utilities',
#     middleware=[CliOutput()]
# )
# class FireflyCli:
#
#     @ffd.cli(
#         target='firefly.Deploy',
#         alias={
#             'env': 'e',
#         },
#         help_={
#             'env': 'The environment to deploy'
#         }
#     )
#     def deploy(self):
#         pass
#
#     # @ffd.cli(target='')
#     # def generate_typescript(self):
#     #     pass



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import logging

import firefly.domain as ffd


class PythonLogger(ffd.Logger):
    _max_length: int = 5120  # ~5k

    def __init__(self):
        self.log = logging
        self.log.basicConfig(format=f'%(message).{self._max_length}s', level=self.log.WARNING)

    def debug(self, message: str, *args, **kwargs):
        self.log.debug(message, *args, **kwargs)

    def info(self, message: str, *args, **kwargs):
        self.log.info(message, *args, **kwargs)

    def warning(self, message: str, *args, **kwargs):
        self.log.warning(message, *args, **kwargs)

    def error(self, message: str, *args, **kwargs):
        self.log.error(message, *args, **kwargs)

    def set_level_to_fatal(self):
        self.log.getLogger().setLevel(logging.FATAL)

    def set_level_to_error(self):
        self.log.getLogger().setLevel(logging.ERROR)

    def set_level_to_warning(self):
        self.log.getLogger().setLevel(logging.WARNING)

    def set_level_to_info(self):
        self.log.getLogger().setLevel(logging.INFO)

    def set_level_to_debug(self):
        self.log.getLogger().setLevel(logging.DEBUG)

    def disable(self):
        self.log.getLogger().setLevel(logging.NOTSET)

    def get_level(self):
        return self.log.getLogger().getEffectiveLevel()

    def set_level(self, level: int):
        self.log.getLogger().setLevel(level)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import List, Union

import firefly as ff
import firefly.domain as ffd
import firefly_di as di
from firefly.domain.entity.core.cli_app import CliApp
from firefly.domain.entity.core.cli_endpoint import CliEndpoint

from ..aggregate_root import AggregateRoot
from ..entity import list_, hidden


class ContextMap(AggregateRoot):
    contexts: List[ffd.Context] = list_()
    _config: ffd.Configuration = hidden()
    _firefly_container: di.Container = hidden()

    def __post_init__(self):
        for name, config in self._config.contexts.items():
            self.contexts.append(ffd.Context(
                name=name,
                config=config or {},
                is_extension=config.get('is_extension', False) if isinstance(config, dict) else False
            ))

        found = False
        for context in self.contexts:
            if context.name == 'firefly':
                found = True
                context.container = self._firefly_container
                break
        if not found:
            c = ffd.Context(name='firefly', config={})
            c.container = self._firefly_container
            self.contexts.append(c)

    def get_context(self, name: str):
        for context in self.contexts:
            if context.name == name:
                return context

    def find_entity_by_name(self, context_name: str, entity_name: str):
        for entity in self.get_context(context_name).entities:
            if entity.__name__ == entity_name:
                return entity

    def locate_service(self, name: str):
        ret = self.locate_command_handler(name)
        if ret is None:
            ret = self.locate_event_listener(name)
        if ret is None:
            ret = self.locate_query_handler(name)

        return ret

    def locate_command_handler(self, command: ff.TypeOfCommand):
        context_name = self._get_context_name_from_message(command)
        context = self.get_context(context_name)

        for command_handler, cmd in context.command_handlers.items():
            if cmd == command:
                return command_handler

    def locate_event_listener(self, event: ff.TypeOfEvent):
        context_name = self._get_context_name_from_message(event)
        context = self.get_context(context_name)

        for event_listener, events in context.command_handlers.items():
            if event in events:
                return event_listener

    def locate_query_handler(self, query: ff.TypeOfQuery):
        context_name = self._get_context_name_from_message(query)
        context = self.get_context(context_name)

        for query_handler, qry in context.command_handlers.items():
            if qry == query:
                return query_handler

    @staticmethod
    def _get_context_name_from_message(message: Union[str, ffd.Message]):
        if isinstance(message, str):
            return message.split('.')[0]
        return message.get_context()

    def get_cli_app(self, name: str):
        app = CliApp(name=name)
        for context in self.contexts:
            for endpoint in context.endpoints:
                if isinstance(endpoint, CliEndpoint) and endpoint.app == name:
                    app.endpoints.append(endpoint)
        return app



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import inspect
import typing


def get_arguments(c: callable, none_type_unions: bool = True) -> dict:
    ret = {}
    sig = dict(inspect.signature(c).parameters)

    for k, v in typing.get_type_hints(c).items():
        if k == 'return':
            continue

        if none_type_unions is False:
            try:
                if len(v.__args__) == 2 and ('NoneType' in str(v.__args__[0]) or 'NoneType' in str(v.__args__[1])):
                    v = v.__args__[0] if 'NoneType' in str(v.__args__[1]) else v.__args__[1]
            except AttributeError:
                pass

        ret[k] = {
            'type': v,
            'default': sig[k].default,
            'kind': sig[k].kind,
        }

    return ret



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import typing

from firefly.domain.meta.context_aware import ContextAware

# __pragma__('skip')
from abc import ABC
# __pragma__('noskip')
# __pragma__ ('ecom')
"""?
from firefly.presentation.web.polyfills import ABC
?"""
# __pragma__ ('noecom')


class FireflyType(ContextAware, ABC):
    _context: str = None

    def __str__(self):
        return f'{self._context}.{self.__class__.__name__}' \
            if self._context is not None else self.__class__.__name__

    def __repr__(self):
        return str(self)

    def is_this(self, this_: typing.Any):
        if isinstance(this_, str) and this_ == str(self):
            return True
        try:
            return isinstance(self, this_) and self.__class__.__name__ == this_.__name__
        except TypeError:
            return False

    def get_context(self):
        return self._context



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import firefly.application as ffa
import firefly.domain as ffd
import firefly.infrastructure as ffi
import firefly_di as di


class Container(di.Container):
    kernel: ffd.Kernel = ffd.Kernel
    event_buffer: ffd.EventBuffer = ffd.EventBuffer
    logger: ffd.Logger = ffi.PythonLogger
    serializer: ffd.Serializer = ffi.JsonSerializer
    json_serializer: ffi.JsonSerializer = ffi.JsonSerializer
    configuration_factory: ffd.ConfigurationFactory = ffi.YamlConfigurationFactory
    configuration: ffd.Configuration = lambda self: self.configuration_factory()
    context_map: ffd.ContextMap = ffd.ContextMap
    registry: ffd.Registry = ffd.Registry
    message_factory: ffd.MessageFactory = ffd.MessageFactory
    validator: ffd.Validation = ffd.Validation

    # System Bus
    event_resolver: ffd.EventResolvingMiddleware = lambda self: self.build(
        ffd.EventResolvingMiddleware, event_listeners={
            ffd.ContainersLoaded: [
                ffa.LoadInfrastructureLayer,
                ffa.LoadDomainLayer,
            ],
            ffd.DomainEntitiesLoaded: [
                ffa.LoadApplicationLayer,
                ffa.AutoGenerateAggregateApis,
            ],
            ffd.ApplicationLayerLoaded: [
                ffa.LoadPresentationLayer,
            ]
        }
    )
    command_resolver: ffd.CommandResolvingMiddleware = lambda self: self.build(
        ffd.CommandResolvingMiddleware, command_handlers={
            ffd.LoadContainers: ffa.LoadContainers,
        }
    )
    query_resolver: ffd.QueryResolvingMiddleware = ffd.QueryResolvingMiddleware
    content_negotiator: ffd.ContentNegotiator = lambda self: ffd.ContentNegotiator({
        'text/html': self.build(ffi.HtmlConverter),
    }, self.logger)
    authenticator: ffa.AuthenticatingMiddleware = ffa.AuthenticatingMiddleware
    transaction_handler: ffd.TransactionHandlingMiddleware = ffd.TransactionHandlingMiddleware
    command_bus: ffd.CommandBus = lambda self: self.build(ffd.CommandBus, middleware=[
        self.build(ffd.LoggingMiddleware),
        self.authenticator,
        self.content_negotiator,
        self.transaction_handler,
        self.build(ffd.EventDispatchingMiddleware),
        self.command_resolver,
    ])
    event_bus: ffd.EventBus = lambda self: self.build(ffd.EventBus, middleware=[
        self.build(ffd.LoggingMiddleware),
        self.authenticator,
        self.transaction_handler,
        self.event_resolver,
    ])
    query_bus: ffd.QueryBus = lambda self: self.build(ffd.QueryBus, middleware=[
        self.build(ffd.LoggingMiddleware),
        self.authenticator,
        self.content_negotiator,
        self.transaction_handler,
        self.query_resolver,
    ])
    system_bus: ffd.SystemBus = ffd.SystemBus

    # Storage
    db_api_storage_interface_registry: ffi.RdbStorageInterfaceRegistry = ffi.RdbStorageInterfaceRegistry

    # Messaging
    message_transport: ffd.MessageTransport = ffi.AsyncioMessageTransport

    # API
    web_server: ffi.WebServer = ffi.WebServer
    cli_executor: ffd.CliAppExecutor = ffi.ArgparseExecutor
    rest_router: ffd.RestRouter = ffi.RoutesRestRouter

    # Deployment
    agent_factory: ffd.AgentFactory = lambda self: self.build(ffd.AgentFactory, agents={
        'default': self.build(ffi.DefaultAgent)
    })



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from firefly.presentation.web.components.layouts.default import menu_item, compose, default_layout
from firefly.presentation.web.js_libs.mithril import m
from firefly.presentation.web.plugins import add_route, add_menu_item
from firefly.presentation.web.polyfills import *  # __:skip

# __pragma__('opov')

m.route.prefix = '/admin'


add_menu_item(m('div.ff-title', 'Kernel'), 0)
add_menu_item(m(menu_item('Configuration', icon='solid/cog')), 1)
add_menu_item(m(menu_item('System Health', icon='solid/heart')), 2)
add_menu_item(m(menu_item('Services', icon='solid/wifi')), 3)

add_route('/', compose(default_layout, lambda: window.ff_menu))

m.route(document.body, '/', window.ff_routes)

"""
__pragma__('js', '{}', '''
if (module.hot) {
  module.hot.accept();
}
''')
"""


#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import List, Type, Dict

import firefly as ff
import firefly.domain as ffd
import firefly_di as di
import inflection

from ..entity import required, hidden, optional, Entity, id_, dict_, list_


class Context(Entity):
    name: str = id_(is_uuid=False)
    config: dict = required()
    container: di.Container = hidden()
    is_extension: bool = optional(default=False)
    entities: List[Type[Entity]] = list_()
    command_handlers: Dict[Type[ffd.ApplicationService], ff.TypeOfCommand] = dict_()
    query_handlers: Dict[Type[ffd.ApplicationService], ff.TypeOfQuery] = dict_()
    event_listeners: Dict[Type[ffd.ApplicationService], List[ff.TypeOfEvent]] = dict_()
    endpoints: List[ff.Endpoint] = list_()

    def __post_init__(self):
        self.ports = []
        self.queue = ffd.Queue(name=f'{inflection.camelize(self.name)}Queue', subscribers=[self])



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import json
from json import JSONDecodeError

import firefly.application as ffa
import firefly.domain as ffd


@ffd.query_handler()
class DefaultHttpMiddleware(ffd.Middleware):
    def __call__(self, message: ffd.Message, next_: callable):
        response = next_(message)

        if message.headers.get('origin') == 'http':
            try:
                return ffd.HttpMessage(http_headers={'status_code': '200', 'content_negotiation-type': 'application/json'},
                                       body=json.dumps(response))
            except (JSONDecodeError, TypeError):
                return response

        return response


@ffd.http(cors=True)
class HttpApi:
    @ffd.http('/services', target=ffa.GetRegisteredContainerServices)
    def services(self):
        pass



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Dict, Optional, Tuple, Union

import firefly.domain as ffd
from routes import Mapper


class RoutesRestRouter(ffd.RestRouter):
    def __init__(self):
        self._maps: Dict[str, Mapper] = {}
        self._cache = {}

    def register(self, route: str, endpoint: ffd.HttpEndpoint):
        if endpoint.method.lower() not in self._maps:
            self._maps[endpoint.method.lower()] = Mapper()
        self._maps[endpoint.method.lower()].connect(route, action=endpoint)
        self._cache[str(endpoint)] = endpoint

    def match(self, route: str, method: str = 'get') -> Union[Tuple[ffd.HttpEndpoint, dict], Tuple[None, None]]:
        result = None
        if method.lower() in self._maps:
            result = self._maps[method.lower()].match(route)
        if result is None and 'any' in self._maps:
            result = self._maps['any'].match(route)

        if result is None:
            return None, None

        action = result['action']
        del result['action']

        return self._cache[action], result



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

import asyncio
import inspect
import logging
import sys
import uuid
from _signal import SIGABRT, SIGILL, SIGINT, SIGSEGV, SIGTERM
from signal import signal
from typing import Callable, Dict
from urllib.parse import parse_qsl

import aiohttp_cors
import firefly.domain as ffd
import websockets
from aiohttp import web
from firefly import TypeOfMessage
from firefly.domain.entity.messaging.http_response import HttpResponse


class WebServer(ffd.SystemBusAware, ffd.LoggerAware):
    _serializer: ffd.Serializer = None
    _message_factory: ffd.MessageFactory = None
    _rest_router: ffd.RestRouter = None

    def __init__(self, host: str = '0.0.0.0', port: int = 9000,
                 websocket_host: str = '0.0.0.0', websocket_port: int = 9001):
        self.routes = []
        self.extensions = []
        self.queues: Dict[str, asyncio.Queue] = {}
        self.queue_map: Dict[str, str] = {}
        self.cors = None
        self.host = host
        self.port = port
        if host.count(':') > 1:
            parts = host.split(':')
            self.port = parts.pop()
            self.host = ':'.join(parts)
        self.websocket_host = websocket_host
        self.websocket_port = websocket_port
        self.app = None
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

    def add_extension(self, extension: Callable):
        self.extensions.append(extension)

    def run(self):
        self.initialize()

        self.loop.run_until_complete(
            websockets.serve(self._handle_websocket, host=self.websocket_host, port=self.websocket_port)
        )

        print(f"Server is running on {self.host}:{self.port}", flush=True)
        web.run_app(self.app, host=self.host, port=self.port)

        self._shut_down()

    def initialize(self):
        self.app = web.Application()
        self._init_logger()

        for extension in self.extensions:
            extension(self)

        self.app.add_routes(self.routes)

        self.cors = aiohttp_cors.setup(self.app, defaults={
            "*": aiohttp_cors.ResourceOptions(
                allow_credentials=True,
                expose_headers="*",
                allow_headers="*",
                allow_methods="*"
            )
        })

        for route in list(self.app.router.routes()):
            self.cors.add(route)

        for sig in (SIGABRT, SIGILL, SIGINT, SIGSEGV, SIGTERM):
            signal(sig, self._shut_down)

        def event_listener(message: ffd.Message, next_: Callable):
            self._broadcast(message)
            return next_(message)
        self._system_bus.add_event_listener(event_listener)

    def _broadcast(self, message: ffd.Message):
        if message.get_context() != 'firefly':
            self.debug('Broadcasting message %s', message)
            m = self._serializer.serialize(message)
            for queue in self.queues.values():
                queue.put_nowait(m)

    def add_endpoint(self, method: str, route: str, message: TypeOfMessage = None):
        print(f'Endpoint: {method} {route} -> {message}')
        self.routes.append(getattr(web, method.lower())(route, self._request_handler_generator(message)))

    def _request_handler_generator(self, msg: TypeOfMessage = None):
        async def _handle_request(request: web.Request):
            self.debug('Got a request -----------------------')
            self.debug(request.headers)
            self.debug(await request.text())
            self.debug('-------------------------------------')

            endpoint, params = self._rest_router.match(request.path, request.method)
            if endpoint is not None:
                if endpoint.message is not None:
                    message_name = endpoint.message if isinstance(endpoint.message, str) else endpoint.message.get_fqn()
                else:
                    message_name = endpoint.service
                    if inspect.isclass(message_name):
                        message_name = message_name.get_fqn()
                if request.method.lower() == 'get':
                    params.update(request.query)
                    message = self._message_factory.query(message_name, None, params)
                else:
                    params.update(self._serializer.deserialize(await request.text()))
                    message = self._message_factory.command(message_name, params)
            else:
                if msg is not None:
                    message_name = msg
                    if inspect.isclass(msg):
                        message_name = message_name.get_fqn()
                    if request.method.lower() == 'get':
                        message = self._message_factory.query(message_name, None, dict(request.query))
                    elif request.method.lower() == 'post':
                        data = await request.text()
                        try:
                            data = dict(self._serializer.deserialize(data))
                        except ffd.InvalidArgument:
                            data = dict(parse_qsl(data))
                        data.update(dict(request.query))
                        message = self._message_factory.command(message_name, data)
                elif request.method.lower() == 'post':
                    message: ffd.Message = self._serializer.deserialize(await request.text())
                else:
                    message: ffd.Message = self._serializer.deserialize(request.query['query'])

                if isinstance(message, dict):
                    endpoint, params = self._rest_router.match(request.path, request.method)
                    if endpoint.message is not None:
                        message_name = endpoint.message if isinstance(endpoint.message, str) else endpoint.message.get_fqn()
                    else:
                        message_name = endpoint.service
                        if inspect.isclass(message_name):
                            message_name = message_name.get_fqn()
                    if request.method.lower() == 'get':
                        message = self._message_factory.query(message_name, None, message)
                    else:
                        message = self._message_factory.command(message_name, message)

            self.debug(f'Decoded message: {message.to_dict()}')

            try:
                message.headers['client_id'] = request.headers['Firefly-Client-ID']
            except KeyError:
                self.info('Request missing header Firefly-ClientID')

            response = None

            await self._marshal_request(message, request)

            status_code = None
            try:
                if isinstance(message, ffd.Event):
                    response = self.dispatch(message)
                elif isinstance(message, ffd.Command):
                    response = self.invoke(message)
                elif isinstance(message, ffd.Query):
                    response = self.request(message)
            except ffd.UnauthenticatedError:
                response = {}
                status_code = 403
            except ffd.UnauthorizedError:
                response = {}
                status_code = 401

            self.debug(f'Response: {response}')

            if isinstance(response, HttpResponse):
                body = response.body
                headers = response.headers
            else:
                body = self._serializer.serialize(response)
                headers = {}

            params = {'body': body, 'headers': headers}
            if status_code:
                params['status'] = status_code

            return web.Response(**params)

        return _handle_request

    @staticmethod
    async def _marshal_request(message: ffd.Message, request: web.Request):
        # TODO make a class
        message.headers['http_request'] = {
            'headers': dict(request.headers),
            'method': request.method,
            'path': request.path,
            'content_type': request.content_type,
            'content_length': request.content_length,
            'query': dict(request.query),
            'post': dict(await request.post()),
            'url': request._message.url,
        }

    async def _handle_websocket(self, websocket, path):
        id_ = str(uuid.uuid4())
        consumer_task = asyncio.ensure_future(self._consumer(websocket, path, id_))
        producer_task = asyncio.ensure_future(self._producer(websocket, path, id_))
        done, pending = await asyncio.wait([consumer_task, producer_task], return_when=asyncio.FIRST_COMPLETED)
        for task in pending:
            task.cancel()

    async def _consumer(self, websocket, path, id_):
        async for m in websocket:
            message = self._serializer.deserialize(m)
            if 'id' in message:
                self.debug(f'Mapping {id_} to client id {message["id"]}')
                self.queue_map[message['id']] = id_

    async def _producer(self, websocket, path, id_):
        self.queues[id_] = asyncio.Queue(loop=self.loop)

        while True:
            m = await self.queues[id_].get()
            print(f'Sending {m}')
            await websocket.send(m)

    @staticmethod
    def _init_logger():
        root = logging.getLogger()
        root.setLevel(logging.DEBUG)

        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        ch.setFormatter(formatter)
        root.addFilter(ch)

    def _shut_down(self):
        print("Shutting down")
        self.loop.stop()
        self.loop.close()



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Union, Callable

import firefly.domain as ffd

from firefly.domain.service.messaging.command_bus import CommandBusAware
from firefly.domain.service.messaging.event_bus import EventBusAware
from firefly.domain.service.messaging.query_bus import QueryBusAware
from firefly.domain.service.messaging.message_bus import MessageBus


def _insert_middleware(bus: MessageBus, which: str, listener: ffd.Middleware, index: int = None, cb: Callable = None,
                       replace: type = None):
    if replace is not None:
        return bus.replace(replace, listener)

    if cb is not None:
        index = cb(which, bus.middleware)
    if index is not None:
        bus.insert(index, listener)
    else:
        bus.add(listener)


class SystemBus(EventBusAware, CommandBusAware, QueryBusAware):
    def add_event_listener(self, listener: ffd.Middleware, **kwargs):
        return _insert_middleware(self._event_bus, 'event', listener, **kwargs)

    def add_command_handler(self, handler: ffd.Middleware, **kwargs):
        return _insert_middleware(self._command_bus, 'command', handler, **kwargs)

    def add_query_handler(self, handler: ffd.Middleware, **kwargs):
        return _insert_middleware(self._query_bus, 'query', handler, **kwargs)

    # deprecated
    def insert_command_handler(self, index: int, handle: ffd.Middleware):
        self._command_bus.insert(index, handle)


class SystemBusAware:
    _system_bus: SystemBus = None

    def dispatch(self, event: Union[ffd.Event, str], data: dict = None):
        return self._system_bus.dispatch(event, data)

    def invoke(self, command: Union[ffd.Command, str], data: dict = None):
        return self._system_bus.invoke(command, data)

    def request(self, request: Union[ffd.Query, str], criteria: Union[ffd.BinaryOp, Callable] = None,
                data: dict = None):
        return self._system_bus.request(request, criteria, data)



#!/usr/bin/env python

#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

""" Unit tests for the ninja.py file. """

import gyp.generator.ninja as ninja
import unittest
import StringIO
import sys
import TestCommon


class TestPrefixesAndSuffixes(unittest.TestCase):
  def test_BinaryNamesWindows(self):
    # These cannot run on non-Windows as they require a VS installation to
    # correctly handle variable expansion.
    if sys.platform.startswith('win'):
      writer = ninja.NinjaWriter('foo', 'wee', '.', '.', 'build.ninja', '.',
          'build.ninja', 'win')
      spec = { 'target_name': 'wee' }
      self.assertTrue(writer.ComputeOutputFileName(spec, 'executable').
          endswith('.exe'))
      self.assertTrue(writer.ComputeOutputFileName(spec, 'shared_library').
          endswith('.dll'))
      self.assertTrue(writer.ComputeOutputFileName(spec, 'static_library').
          endswith('.lib'))

  def test_BinaryNamesLinux(self):
    writer = ninja.NinjaWriter('foo', 'wee', '.', '.', 'build.ninja', '.',
        'build.ninja', 'linux')
    spec = { 'target_name': 'wee' }
    self.assertTrue('.' not in writer.ComputeOutputFileName(spec,
                                                            'executable'))
    self.assertTrue(writer.ComputeOutputFileName(spec, 'shared_library').
        startswith('lib'))
    self.assertTrue(writer.ComputeOutputFileName(spec, 'static_library').
        startswith('lib'))
    self.assertTrue(writer.ComputeOutputFileName(spec, 'shared_library').
        endswith('.so'))
    self.assertTrue(writer.ComputeOutputFileName(spec, 'static_library').
        endswith('.a'))

if __name__ == '__main__':
  unittest.main()



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from firefly.presentation.web.js_libs.mithril import m
from firefly.presentation.web.polyfills import *  # __:skip

# __pragma__('kwargs')

os = None
agent = navigator.userAgent or navigator.vendor or window.opera
if 'windows phone' in agent:
    os = 'windows'
elif 'android' in agent:
    os = 'android'
elif 'iPad' in agent or 'iPhone' in agent or 'iPod' in agent:
    os = 'ios'


class Icon:
    def __init__(self, name: str, onclick=None):
        self._name = name
        self._onclick = onclick

    def view(self):
        icon = None
        name = self._name
        # __pragma__('js', '{}', "icon = require(`@fortawesome/fontawesome-free/svgs/${py_name}.svg`);")

        if self._onclick is not None:
            return m('span.cursor-pointer', {'onclick': self._onclick}, m.trust(icon))

        return m.trust(icon)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import firefly.domain as ffd
import firefly_di as di

from .application_service import ApplicationService


class AddApplicationService(ApplicationService):
    _container: di.Container = None
    _event_resolver: ffd.EventResolvingMiddleware = None
    _command_resolver = ffd.CommandResolvingMiddleware = None
    _query_resolver = ffd.QueryResolvingMiddleware = None

    def __call__(self, fqn: str, args: dict = None):
        args = args or {}
        cls = ffd.load_class(fqn)
        for key in ('__ff_command_handler', '__ff_listener', '__ff_query_handler'):
            if hasattr(cls, key):
                mw = ffd.ServiceExecutingMiddleware(self._container.build(cls, **args))
                setattr(mw, key, getattr(cls, key))
                self._add_middleware(mw)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Callable, TypeVar, Type

import firefly.domain as ffd

from firefly.domain.service.messaging.middleware import Middleware
from firefly.domain.entity.messaging.message import Message

M = TypeVar('M', bound=Message)


class SubscriptionWrapper(Middleware):
    def __init__(self, middleware: ffd.Middleware, type_: Type[M]):
        self._middleware = middleware
        self._type = type_

    def __call__(self, message: ffd.Message, next_: Callable) -> ffd.Message:
        if (isinstance(self._type, str) and message == self._type) or (
                not isinstance(self._type, str) and isinstance(message, self._type)):
            return self._middleware(message, next_)
        else:
            return next_(message)

    def __repr__(self):
        return '<SubscriptionWrapper {}, {}>'.format(repr(self._middleware), repr(self._type))



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from dataclasses import is_dataclass
from datetime import datetime, date
from typing import Union, List, Type

import firefly.domain as ffd


class EntityAttributeSpy:
    def __init__(self, type_: Type[ffd.ValueObject] = None):
        self._type = type_

    def __getattribute__(self, item):
        if object.__getattribute__(self, '_type') is not None:
            t = object.__getattribute__(self, '_type')
            valid = False
            if is_dataclass(t) and item in t.__dataclass_fields__:
                valid = True
            elif hasattr(t, item):
                valid = True
            if not valid:
                raise AttributeError(f"'{t.__name__}' object has no attribute '{item}'")
        return Attr(item)


class AttributeString(str):
    pass


class Invalid:
    pass


INVALID = Invalid()


class Attr:
    def __init__(self, attr: str, default=None):
        self.attr = AttributeString(attr)
        self.default = default

    def __getattribute__(self, item):
        try:
            return object.__getattribute__(self, item)
        except AttributeError:
            raise ffd.InvalidOperand(f"Use of '{item}' is not currently supported.")

    def is_none(self):
        return BinaryOp(self.attr, 'is', 'None')

    def is_false(self):
        return BinaryOp(self.attr, 'is', False)

    def is_true(self):
        return BinaryOp(self.attr, 'is', True)

    def is_in(self, set_):
        return BinaryOp(self.attr, 'in', set_)

    def contains(self, value):
        return BinaryOp(self.attr, 'contains', value)

    def startswith(self, value):
        return BinaryOp(self.attr, 'startswith', value)

    def endswith(self, value):
        return BinaryOp(self.attr, 'endswith', value)

    def __eq__(self, other):
        return BinaryOp(self.attr, '==', other)

    def __ne__(self, other):
        return BinaryOp(self.attr, '!=', other)

    def __gt__(self, other):
        return BinaryOp(self.attr, '>', other)

    def __ge__(self, other):
        return BinaryOp(self.attr, '>=', other)

    def __lt__(self, other):
        return BinaryOp(self.attr, '<', other)

    def __le__(self, other):
        return BinaryOp(self.attr, '<=', other)

    def __repr__(self):
        return self.attr


class AttrFactory:
    def __init__(self, fields: List[str]):
        self._fields = fields

    def __getattr__(self, item):
        if item in self._fields:
            return Attr(item)
        return object.__getattribute__(self, item)


class BinaryOp:
    def __init__(self, lhv, op, rhv):
        self.lhv = lhv
        self.op = op
        self.rhv = rhv

    def to_dict(self):
        return self._do_to_dict(self)

    def _do_to_dict(self, bop: BinaryOp):
        ret = {'l': None, 'o': bop.op, 'r': None}

        if isinstance(bop.lhv, BinaryOp):
            ret['l'] = self._do_to_dict(bop.lhv)
        elif isinstance(bop.lhv, (Attr, AttributeString)):
            ret['l'] = f'a:{str(bop.lhv)}'
        else:
            ret['l'] = bop.lhv

        if isinstance(bop.rhv, BinaryOp):
            ret['r'] = self._do_to_dict(bop.rhv)
        elif isinstance(bop.rhv, (Attr, AttributeString)):
            ret['r'] = f'a:{str(bop.rhv)}'
        else:
            ret['r'] = bop.rhv

        return ret

    @classmethod
    def from_dict(cls, data: dict):
        if isinstance(data['l'], dict):
            lhv = cls.from_dict(data['l'])
        elif isinstance(data['l'], str) and data['l'].startswith('a:'):
            lhv = Attr(data['l'].split(':')[1])
        else:
            lhv = data['l']

        if isinstance(data['r'], dict):
            rhv = cls.from_dict(data['r'])
        elif isinstance(data['r'], str) and data['r'].startswith('a:'):
            rhv = Attr(data['r'].split(':')[1])
        else:
            rhv = data['r']

        return BinaryOp(lhv, data['o'], rhv)

    def matches(self, data: Union[ffd.Entity, dict]) -> bool:
        if isinstance(data, ffd.Entity):
            data = data.to_dict()

        return self._do_match(self, data)

    def _do_match(self, bop: BinaryOp, data: dict) -> bool:
        if isinstance(bop.lhv, BinaryOp):
            lhv = self._do_match(bop.lhv, data)
        elif isinstance(bop.lhv, AttributeString):
            lhv = data[bop.lhv]
        elif isinstance(bop.lhv, Attr):
            lhv = data[bop.lhv.attr]
        else:
            lhv = bop.lhv

        if isinstance(bop.rhv, BinaryOp):
            rhv = self._do_match(bop.rhv, data)
        elif isinstance(bop.rhv, AttributeString):
            rhv = data[bop.rhv]
        elif isinstance(bop.rhv, Attr):
            rhv = data[bop.rhv.attr]
        else:
            rhv = bop.rhv

        if bop.op == '==':
            return lhv == rhv
        if bop.op == '!=':
            return lhv != rhv
        if bop.op == '>':
            return lhv > rhv
        if bop.op == '<':
            return lhv < rhv
        if bop.op == '>=':
            return lhv >= rhv
        if bop.op == '<=':
            return lhv <= rhv
        if bop.op == 'is':
            return lhv is rhv if rhv != 'None' else lhv is None
        if bop.op == 'in':
            return lhv in rhv
        if bop.op == 'contains':
            return rhv in lhv
        if bop.op == 'startswith':
            return str(lhv).startswith(rhv)
        if bop.op == 'endswith':
            return str(lhv).endswith(rhv)
        if bop.op == 'and':
            return lhv and rhv
        if bop.op == 'or':
            return lhv or rhv

        raise ffd.LogicError(f"Don't know how to handle op: {bop.op}")

    def __and__(self, other):
        return BinaryOp(self, 'and', other)

    def __or__(self, other):
        return BinaryOp(self, 'or', other)

    def prune(self, fields: list):
        data = self._prune(fields, self.to_dict())
        if data is INVALID:
            data = {'l': 1, 'o': '==', 'r': 1}
        return BinaryOp.from_dict(data)

    def _prune(self, fields: list, data: dict):
        if isinstance(data['l'], dict):
            result = self._prune(fields, data['l'])
            if result is INVALID:
                if data['o'] == 'and':
                    data['l'] = {'l': 1, 'o': '==', 'r': 1}
                else:
                    data['l'] = {'l': 1, 'o': '!=', 'r': 1}
        elif isinstance(data['l'], str) and len(data['l']) > 2 and \
                data['l'].startswith('a:') and data['l'][2:] not in fields:
            return INVALID

        if isinstance(data['r'], dict):
            result = self._prune(fields, data['r'])
            if result is INVALID:
                if data['o'] == 'and':
                    data['r'] = {'l': 1, 'o': '==', 'r': 1}
                else:
                    data['r'] = {'l': 1, 'o': '!=', 'r': 1}
        elif isinstance(data['r'], str) and len(data['r']) > 2 and \
                data['r'].startswith('a:') and data['r'][2:] not in fields:
            return INVALID

        if isinstance(data['r'], dict) and isinstance(data['l'], dict):
            if data['l']['l'] == 1 and data['l']['r'] == 1 and data['r']['l'] == 1 and data['r']['r'] == 1:
                return INVALID

        return data

    def to_sql(self):
        sql, params, counter = self._to_sql()
        return sql, params

    def _to_sql(self, counter: int = None, params: dict = None):
        counter = counter or 1
        params = params or {}
        lhv, params, counter = self._process_op(self.lhv, params, counter)
        rhv, params, counter = self._process_op(self.rhv, params, counter)

        return f'({lhv} {self.op.replace("==", "=")} {rhv})', params, counter

    @staticmethod
    def _process_op(v, params: dict, counter: int):
        if isinstance(v, BinaryOp):
            v, p, counter = v._to_sql(counter, params)
            params.update(p)
        elif isinstance(v, (list, tuple)):
            placeholders = []
            for i in v:
                var = f'var{counter}'
                counter += 1
                placeholders.append(f':{var}')
                params[var] = i
            v = f'({",".join(placeholders)})'
        elif not isinstance(v, (Attr, AttributeString)):
            var = f'var{counter}'
            counter += 1
            params[var] = v
            v = f':{var}'

        return v, params, counter

    def __repr__(self):
        return f'({self.lhv} {self.op} {self.rhv})'

    def __eq__(self, other):
        return isinstance(other, BinaryOp) and self.to_dict() == other.to_dict()



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

# __pragma__('skip')
from typing import Tuple, List, Union, Type

from .domain import *

EventList = Union[Event, Tuple[str, Union[dict, object]], List[Union[Event, Tuple[str, Union[dict, object]]]]]
TypeOfMessage = Union[str, Type[Message]]
TypeOfCommand = Union[str, Type[Command]]
TypeOfEvent = Union[str, Type[Event]]
TypeOfQuery = Union[str, Type[Query]]
# __pragma__('noskip')
# __pragma__ ('ecom')
"""?
from firefly.presentation.web.polyfills import Entity, AggregateRoot, Message, Command, Query, Event, ValueObject
from firefly.domain.entity.entity import required, optional, id_, now, list_, dict_, today, hidden
from firefly.domain.entity.validation.validators import IsType, HasLength, Matches, IsValidEmail, IsOneOf, IsDatetime, IsNumeric, IsFloat, IsInt, IsValidUrl
from firefly.domain.error import MissingArgument, FrameworkError
from firefly.domain.service.entity.validator import Validator
from firefly.domain.service.messaging import SystemBus, Middleware, CommandBus, EventBus, QueryBus, MessageFactory
?"""
# __pragma__ ('noecom')



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import List, Callable, Optional, Union

import firefly.domain as ffd
import firefly.infrastructure as ffi
import inflection
from firefly.domain.repository.repository import T


class RdbRepository(ffd.Repository[T]):
    def __init__(self, interface: ffi.RdbStorageInterface, table_name: str = None):
        super().__init__()

        self._entity_type = self._type()
        self._table = table_name or inflection.tableize(self._entity_type.get_fqn())
        self._interface = interface
        self._index = 0
        self._state = 'empty'

    def execute_ddl(self):
        self._interface.execute_ddl(self._type())

    def append(self, entity: T):
        self.debug('Entity added to repository: %s', str(entity))
        if entity not in self._entities:
            self._entities.append(entity)
        self._state = 'partial'

    def remove(self, entity: T):
        self.debug('Entity removed from repository: %s', str(entity))
        self._deletions.append(entity)
        self._entities.remove(entity)

    def find(self, exp: Union[str, Callable]) -> T:
        ret = None
        if isinstance(exp, str):
            entity = self._find_checked_out_entity(exp)
            if entity is not None:
                return entity
            ret = self._interface.find(exp, self._entity_type)
        else:
            results = self._interface.all(self._entity_type, self._get_search_criteria(exp))
            if len(results) > 0:
                ret = results[0]

        if ret:
            self._register_entity(ret)
            if self._state == 'empty':
                self._state = 'partial'

        return ret

    def raw(self, cb: Union[Callable, ffd.BinaryOp] = None, limit: int = None):
        criteria = None
        if cb is not None:
            criteria = self._get_search_criteria(cb)
        return self._interface.raw(self._entity_type, criteria, limit)

    def filter(self, cb: Union[Callable, ffd.BinaryOp]) -> List[T]:
        if self._state == 'full':
            criteria = self._get_search_criteria(cb)
            entities = list(filter(lambda e: criteria.matches(e), self._entities))
        else:
            criteria = self._get_search_criteria(cb)
            entities = self._interface.all(self._entity_type, criteria=criteria)

            for entity in entities:
                self._register_entity(entity)
            if self._state == 'empty':
                self._state = 'partial'
        return entities

    def reduce(self, cb: Callable) -> Optional[T]:
        pass

    def __iter__(self):
        self._load_all()
        return iter(list(self._entities))

    def __next__(self):
        pass

    def __len__(self):
        self._load_all()
        return len(self._entities)

    def __getitem__(self, item):
        self._load_all()
        return self._entities[item]

    def _load_all(self):
        if self._state != 'full':
            for entity in self._interface.all(self._entity_type):
                if entity not in self._entities:
                    self._register_entity(entity)
            self._state = 'full'

    def commit(self):
        self.debug('commit() called in %s', str(self))
        for entity in self._deletions:
            self.debug('Deleting %s', entity)
            self._interface.remove(entity)

        for entity in self._new_entities():
            self.debug('Adding %s', entity)
            self._interface.add(entity)

        for entity in self._changed_entities():
            self.debug('Updating %s', entity)
            self._interface.update(entity)
        self.debug('Done in commit()')

    def __repr__(self):
        return f'DbApiRepository[{self._entity_type}]'

    def _find_checked_out_entity(self, id_: str):
        for entity in self._entities:
            if entity.id_value() == id_:
                return entity

    def reset(self):
        super().reset()
        self._state = 'empty'



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from abc import ABC
from typing import List, Dict, Type, Optional

import firefly as ff


class MetaAware(ABC):
    _events: Dict[Type[MetaAware], List[ff.TypeOfEvent]] = {}
    _command: Dict[Type[MetaAware], ff.TypeOfCommand] = {}
    _query: Dict[Type[MetaAware], ff.TypeOfQuery] = {}
    _endpoints: Dict[Type[MetaAware], List[ff.Endpoint]] = {}
    _agent: Optional[str] = None
    _middleware_config: Optional[Dict] = None
    _annotations: Dict[Type[MetaAware], List[ff.ConfigurationAnnotation]] = {}

    @classmethod
    def is_handler(cls):
        return cls.is_event_listener() or cls.is_query_handler() or cls.is_command_handler()

    @classmethod
    def is_middleware(cls):
        return cls._middleware_config is not None

    @classmethod
    def add_event(cls, event: ff.TypeOfEvent):
        if cls not in cls._events:
            cls._events[cls] = []
        cls._events[cls].append(event)

    @classmethod
    def add_endpoint(cls, endpoint: ff.Endpoint):
        if cls not in cls._endpoints:
            cls._endpoints[cls] = []
        cls._endpoints[cls].append(endpoint)

    @classmethod
    def add_annotation(cls, annotation: ff.ConfigurationAnnotation):
        if cls not in cls._annotations:
            cls._annotations[cls] = []
        cls._annotations[cls].append(annotation)

    @classmethod
    def set_command(cls, command: ff.TypeOfCommand):
        cls._command[cls] = command

    @classmethod
    def set_query(cls, query: ff.TypeOfQuery):
        cls._query[cls] = query

    @classmethod
    def is_event_listener(cls):
        return cls in cls._events

    @classmethod
    def is_command_handler(cls):
        return cls in cls._command

    @classmethod
    def is_query_handler(cls):
        return cls in cls._query

    @classmethod
    def has_endpoints(cls):
        return cls in cls._endpoints

    @classmethod
    def has_annotations(cls):
        return cls in cls._annotations

    @classmethod
    def get_events(cls):
        return cls._events[cls]

    @classmethod
    def get_command(cls):
        return cls._command[cls]

    @classmethod
    def get_query(cls):
        return cls._query[cls]

    @classmethod
    def get_endpoints(cls):
        return cls._endpoints[cls]

    @classmethod
    def get_annotations(cls):
        return cls._annotations[cls]

    @classmethod
    def is_agent(cls):
        return cls._agent is not None

    @classmethod
    def set_agent(cls, agent: str):
        cls._agent = agent

    @classmethod
    def get_agent(cls):
        return cls._agent

    @classmethod
    def set_middleware_config(cls, config: dict):
        cls._middleware_config = config

    @classmethod
    def get_middleware_config(cls):
        return cls._middleware_config



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import TypeVar, Generic, Optional, Union

import firefly.domain as ffd

from .crud_operation import CrudOperation
from ..core.application_service import ApplicationService
from ..messaging.system_bus import SystemBusAware
from ...value_object.generic_base import GenericBase

T = TypeVar('T')


class UpdateEntity(Generic[T], ApplicationService, GenericBase, CrudOperation, SystemBusAware):
    _registry: ffd.Registry = None

    def __call__(self, **kwargs) -> bool:
        type_ = self._type()
        id_arg = type_.match_id_from_argument_list(kwargs)
        entity = self._registry(type_).find(list(id_arg.values()).pop())
        for k, v in kwargs.items():
            if hasattr(entity, k):
                setattr(entity, k, v)
        self.dispatch(self._build_event(type_, 'update'))

        return True



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

"""Visual Studio user preferences file writer."""

import os
import re
import socket # for gethostname

import gyp.common
import gyp.easy_xml as easy_xml


#------------------------------------------------------------------------------

def _FindCommandInPath(command):
  """If there are no slashes in the command given, this function
     searches the PATH env to find the given command, and converts it
     to an absolute path.  We have to do this because MSVS is looking
     for an actual file to launch a debugger on, not just a command
     line.  Note that this happens at GYP time, so anything needing to
     be built needs to have a full path."""
  if '/' in command or '\\' in command:
    # If the command already has path elements (either relative or
    # absolute), then assume it is constructed properly.
    return command
  else:
    # Search through the path list and find an existing file that
    # we can access.
    paths = os.environ.get('PATH','').split(os.pathsep)
    for path in paths:
      item = os.path.join(path, command)
      if os.path.isfile(item) and os.access(item, os.X_OK):
        return item
  return command

def _QuoteWin32CommandLineArgs(args):
  new_args = []
  for arg in args:
    # Replace all double-quotes with double-double-quotes to escape
    # them for cmd shell, and then quote the whole thing if there
    # are any.
    if arg.find('"') != -1:
      arg = '""'.join(arg.split('"'))
      arg = '"%s"' % arg

    # Otherwise, if there are any spaces, quote the whole arg.
    elif re.search(r'[ \t\n]', arg):
      arg = '"%s"' % arg
    new_args.append(arg)
  return new_args

class Writer(object):
  """Visual Studio XML user user file writer."""

  def __init__(self, user_file_path, version, name):
    """Initializes the user file.

    Args:
      user_file_path: Path to the user file.
      version: Version info.
      name: Name of the user file.
    """
    self.user_file_path = user_file_path
    self.version = version
    self.name = name
    self.configurations = {}

  def AddConfig(self, name):
    """Adds a configuration to the project.

    Args:
      name: Configuration name.
    """
    self.configurations[name] = ['Configuration', {'Name': name}]

  def AddDebugSettings(self, config_name, command, environment = {},
                       working_directory=""):
    """Adds a DebugSettings node to the user file for a particular config.

    Args:
      command: command line to run.  First element in the list is the
        executable.  All elements of the command will be quoted if
        necessary.
      working_directory: other files which may trigger the rule. (optional)
    """
    command = _QuoteWin32CommandLineArgs(command)

    abs_command = _FindCommandInPath(command[0])

    if environment and isinstance(environment, dict):
      env_list = ['%s="%s"' % (key, val)
                  for (key,val) in environment.iteritems()]
      environment = ' '.join(env_list)
    else:
      environment = ''

    n_cmd = ['DebugSettings',
             {'Command': abs_command,
              'WorkingDirectory': working_directory,
              'CommandArguments': " ".join(command[1:]),
              'RemoteMachine': socket.gethostname(),
              'Environment': environment,
              'EnvironmentMerge': 'true',
              # Currently these are all "dummy" values that we're just setting
              # in the default manner that MSVS does it.  We could use some of
              # these to add additional capabilities, I suppose, but they might
              # not have parity with other platforms then.
              'Attach': 'false',
              'DebuggerType': '3',  # 'auto' debugger
              'Remote': '1',
              'RemoteCommand': '',
              'HttpUrl': '',
              'PDBPath': '',
              'SQLDebugging': '',
              'DebuggerFlavor': '0',
              'MPIRunCommand': '',
              'MPIRunArguments': '',
              'MPIRunWorkingDirectory': '',
              'ApplicationCommand': '',
              'ApplicationArguments': '',
              'ShimCommand': '',
              'MPIAcceptMode': '',
              'MPIAcceptFilter': ''
             }]

    # Find the config, and add it if it doesn't exist.
    if config_name not in self.configurations:
      self.AddConfig(config_name)

    # Add the DebugSettings onto the appropriate config.
    self.configurations[config_name].append(n_cmd)

  def WriteIfChanged(self):
    """Writes the user file."""
    configs = ['Configurations']
    for config, spec in sorted(self.configurations.iteritems()):
      configs.append(spec)

    content = ['VisualStudioUserFile',
               {'Version': self.version.ProjectVersion(),
                'Name': self.name
               },
               configs]
    easy_xml.WriteXmlIfChanged(content, self.user_file_path,
                               encoding="Windows-1252")



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import argparse
import inspect
import logging
from argparse import ArgumentParser
from typing import Type

import firefly.domain as ffd
import inflection


class App:
    def __init__(self, context_map: ffd.ContextMap, system_bus: ffd.SystemBus, message_factory: ffd.MessageFactory,
                 name: str):
        self.name = name
        self._message_factory = message_factory
        self._context_map = context_map
        self._bus = system_bus
        self._ports = []
        self._parser = ArgumentParser()
        self._subparsers = None
        self._message_lookup = {}

        self._initialize()

    def run(self):
        args = self._parser.parse_args()

        if args.verbose is True:
            logging.getLogger().setLevel(logging.INFO)
        if args.debug is True:
            logging.getLogger().setLevel(logging.DEBUG)

        try:
            message = self._message_lookup[args.target]['target']
        except KeyError:
            self._parser.print_usage()
            return

        if isinstance(message, str):
            service = self._context_map.locate_service(message)
            param_target = service.__call__
            _, type_ = service.locate_message(message)
        else:
            param_target = message
            _, type_ = message.locate_message(message)

        result = self._dispatch_message(message, args, param_target, type_)

        handle = ffd.MiddlewareStack(self._message_lookup[args.target]['middleware'])
        result = handle(result)

        print(result)

    def _dispatch_message(self, message, args, target, type_: str):
        if isinstance(message, str):
            message = getattr(self._message_factory, type_)(message, ffd.build_argument_list(args.__dict__, target))
        else:
            message = message(ffd.build_argument_list(args.__dict__, target))

        if type_ == 'command':
            return self._bus.invoke(message)
        if type_ == 'event':
            return self._bus.dispatch(message)
        return self._bus.query(message)

    def _initialize(self):
        self._gather_ports()
        self._add_verbosity_arguments(self._parser)
        self._add_commands()

    def _gather_ports(self):
        for context in self._context_map.contexts:
            for port in context.ports:
                if not hasattr(port, '__ff_cli'):
                    continue
                config = getattr(port, '__ff_cli')
                if config['app_name'] != self.name:
                    continue
                self._ports.append(port)

    def _get_subparsers(self):
        if self._subparsers is None:
            self._subparsers = self._parser.add_subparsers(dest='target')
        return self._subparsers

    @staticmethod
    def _add_verbosity_arguments(parser):
        try:
            parser.add_argument('--verbose', '-v', action='store_true', default=False, help='Enable verbose output')
            parser.add_argument('--debug', '-d', action='store_true', default=False, help='Enable debug output')
        except argparse.ArgumentError:
            pass

    def _add_commands(self, ports: list = None, options: dict = None, parser=None, parent_class=None):
        top_level = False
        if ports is None:
            top_level = True
            ports = self._ports
        if options is None:
            options = {'alias': {}, 'help_': {}, 'middleware': [], 'params': {}}
        if parser is None:
            parser = self._parser

        for port in ports:
            config = getattr(port, '__ff_cli')
            if inspect.isclass(port):
                if not top_level:
                    parser = self._get_subparsers().add_parser(config['name'], help=config['description'])
                options['alias'].update(config['alias'] if config['alias'] is not None else {})
                options['help_'].update(config['help_'] if config['help_'] is not None else {})
                options['middleware'].extend(config['middleware'] if config['middleware'] is not None else [])
                if hasattr(port, '__init__'):
                    options['params'].update(ffd.get_arguments(port.__init__))
                class_ports = []
                for k, v in port.__dict__.items():
                    if hasattr(v, '__ff_cli'):
                        class_ports.append(v)
                self._add_commands(ports=class_ports, options=options, parser=parser, parent_class=port)
            elif inspect.isfunction(port):
                name = inflection.dasherize(config['name'])
                parser = self._get_subparsers().add_parser(name, help=config['description'])
                mw = options['middleware'].copy()
                sig = inspect.signature(port)
                if 'message' in sig.parameters and 'next_' in sig.parameters:
                    instance = parent_class()
                    mw.append(getattr(instance, port.__name__))

                self._message_lookup[name] = {
                    'target': config['target'],
                    'middleware': mw,
                }
                self._add_adapter(port, config, options, parser)

    def _add_adapter(self, func, config: dict, options: dict, parser):
        message = config['target']
        params = None
        if isinstance(message, str):
            service = self._context_map.locate_service(message)
            if service is not None:
                params = ffd.get_arguments(service.__call__)
        else:
            params = ffd.get_arguments(message)

        if params is None:
            raise ffd.FrameworkError(f'Could not locate service {message}')

        options['params'].update(params)
        options['alias'].update(config['alias'] if config['alias'] is not None else {})
        options['help_'].update(config['help_'] if config['help_'] is not None else {})
        options['middleware'].extend(config['middleware'] if config['middleware'] is not None else [])
        for name, config in options['params'].items():
            args = [f'--{inflection.dasherize(name)}']
            if name in options['alias']:
                args.append(f'-{options["alias"][name]}')

            kwargs = {
                'default': config['default'],
                'required': config['default'] is inspect.Parameter.empty,
                'help': options['help_'][name] if name in options['help_'] else None,
            }

            if config['type'] == bool:
                kwargs['action'] = 'store_true' if not config['default'] else 'store_false'
            else:
                kwargs['type'] = config['type']

            parser.add_argument(*args, **kwargs)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from firefly.domain.entity.entity import Entity

# __pragma__('skip')
from dataclasses import Field
from typing import Type
# __pragma__('noskip')
# __pragma__ ('ecom')
"""?
from firefly.presentation.web.polyfills import Field, Type
?"""
# __pragma__ ('noecom')


class Validator:
    def validate(self, dto: dict, against: Type[Entity]):
        validations = {}
        for k in dir(against):
            v = getattr(against, k)
            if isinstance(v, Field):
                validators = self._parse_field(v)
            elif isinstance(v, dict):
                validators = self._parse_dict(v)
            else:
                continue
            if validators['validators'] is not None:
                validations[k] = validators

        error_messages = {}
        count = 0
        for k, validators in validations.items():
            if (k not in dto or dto[k] is None) and not validators['required']:
                continue

            for validator in validators['validators']:
                if not getattr(validator, '__call__')(dto[k], dto):
                    if k not in error_messages:
                        error_messages[k] = []
                    error_messages[k].append(validator.message.format(k))
                    count += 1

        return {
            'count': count,
            'errors': error_messages,
        }

    @staticmethod
    def _parse_field(value: Field):
        ret = {
            'required': True,
            'validators': None
        }

        if 'required' in value.metadata and not value.metadata['required']:
            ret['required'] = False

        if 'validators' in value.metadata:
            ret['validators'] = value.metadata['validators']

        return ret

    @staticmethod
    def _parse_dict(value: dict):
        ret = {
            'required': True,
            'validators': None
        }

        if isinstance(value, dict) and 'metadata' in value:
            if 'required' in value['metadata'] and not value['metadata']['required']:
                ret['required'] = False

            if 'validators' in value['metadata']:
                ret['validators'] = value['metadata']['validators']

        return ret



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from dataclasses import is_dataclass, asdict
from typing import Callable

import firefly.domain as ffd

from firefly.domain.service.messaging.middleware import Middleware
from firefly.domain.service.messaging.system_bus import SystemBusAware


class EventDispatchingMiddleware(Middleware, SystemBusAware):
    _event_buffer: ffd.EventBuffer = None
    _message_factory: ffd.MessageFactory = None

    def __call__(self, message: ffd.Message, next_: Callable) -> ffd.Message:
        try:
            ret = next_(message)
            for event in self._event_buffer:
                data = event[1]
                if is_dataclass(data):
                    data = asdict(data)
                if isinstance(event, tuple):
                    self.dispatch(self._message_factory.event(event[0], data))
                else:
                    self.dispatch(event)
        finally:
            self._event_buffer.clear()

        return ret



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import inspect
import json
from datetime import datetime, date
from json import JSONEncoder
from pprint import pprint

import firefly.domain as ffd
import firefly_di as di


class FireflyEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, ffd.Entity):
            return o.to_dict()
        elif inspect.isclass(o) and issubclass(o, ffd.Entity):
            return f'{o.get_class_context()}.{o.__name__}'
        elif isinstance(o, (di.Container, ffd.Empty)):
            return None
        elif isinstance(o, (datetime, date)):
            return o.isoformat()
        elif not inspect.isclass(o) and isinstance(o, ffd.Message):
            dic = o.to_dict()
            dic['_name'] = o.__class__.__name__
            t = 'event'
            if isinstance(o, ffd.Command):
                t = 'command'
            elif isinstance(o, ffd.Query):
                t = 'query'
            dic['_type'] = t
            return dic

        return JSONEncoder.default(self, o)


class JsonSerializer(ffd.Serializer):
    _message_factory: ffd.MessageFactory = None

    def serialize(self, data):
        return json.dumps(data, cls=FireflyEncoder, skipkeys=True)

    def deserialize(self, data):
        try:
            if isinstance(data, (str, bytes, bytearray)):
                ret = json.loads(data)
            else:
                ret = data
        except json.JSONDecodeError:
            raise ffd.InvalidArgument('Could not deserialize data')

        if isinstance(ret, dict) and '_name' in ret:
            fqn = f'{ret["_context"]}.{ret["_name"]}'
            t = ffd.load_class(fqn)
            if t is None:
                args = [fqn]
                if ret['_type'] == 'query':
                    args.append(None)
                args.append(ret)
                return getattr(self._message_factory, ret['_type'])(*args)
            return t(**ffd.build_argument_list(ret, t))

        return ret



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import sqlite3
from typing import Type, Optional

import firefly.domain as ffd
import inflection

from ..rdb_storage_interface import RdbStorageInterface


class SqliteStorageInterface(RdbStorageInterface, ffd.LoggerAware):
    _serializer: ffd.Serializer = None

    def __init__(self, **kwargs):
        super().__init__()
        self._config = kwargs
        self._connection: Optional[sqlite3.Connection] = None

    def _disconnect(self):
        if self._connection is not None:
            self._connection.close()
            self._connection = None
            self._cache = {}
            self._tables_checked = []

    def _add(self, entity: ffd.Entity):
        cursor = self._connection.cursor()
        cursor.execute(*self._generate_insert(entity))
        self._connection.commit()

    def _all(self, entity_type: Type[ffd.Entity], criteria: ffd.BinaryOp = None, limit: int = None):
        cursor = self._connection.cursor()
        sql = f"select obj from {self._fqtn(entity_type)}"
        params = {}

        pruned_criteria = None
        if criteria is not None:
            pruned_criteria = criteria.prune([i.name for i in self.get_indexes(entity_type)])
            clause, params = self._generate_where_clause(pruned_criteria)
            if len(clause) > 0:
                sql = f'{sql} where {clause}'
                self.debug('Searching: %s. Params: %s', sql, params)
        cursor.execute(sql, params)

        ret = []
        row = cursor.fetchone()
        while row is not None:
            data = self._serializer.deserialize(row['obj'])
            e = entity_type.from_dict(data)
            ret.append(e)
            if limit is not None and len(ret) >= limit:
                break
            row = cursor.fetchone()

        if limit == 1 and len(ret) > 0:
            return ret[0]

        if criteria != pruned_criteria:
            ret = list(filter(lambda ee: criteria.matches(ee), ret))

        return ret

    def _find(self, uuid: str, entity_type: Type[ffd.Entity]):
        cursor = self._connection.cursor()
        sql = f"select obj from {self._fqtn(entity_type)} where id = ?"
        cursor.execute(sql, (uuid,))
        row = cursor.fetchone()
        if row is not None:
            return entity_type.from_dict(self._serializer.deserialize(row['obj']))

    def _remove(self, entity: ffd.Entity):
        cursor = self._connection.cursor()
        sql = f"delete from {self._fqtn(entity.__class__)} where id = ?"
        cursor.execute(sql, (entity.id_value(),))
        self._connection.commit()
        self._remove_from_cache(entity)

    def _update(self, entity: ffd.Entity):
        cursor = self._connection.cursor()
        cursor.execute(*self._generate_update(entity))
        self._connection.commit()

    def _ensure_connected(self):
        if self._connection is not None:
            return

        try:
            host = self._config['host']
        except KeyError:
            raise ffd.ConfigurationError(f'host is required in sqlite_storage_interface')

        self._connection = sqlite3.connect(host, detect_types=sqlite3.PARSE_DECLTYPES)
        self._connection.row_factory = sqlite3.Row

    def _execute_ddl(self, entity: Type[ffd.Entity]):
        self._ensure_connected()
        cursor = self._connection.cursor()
        self.debug(self._generate_create_table(entity))
        cursor.execute(self._generate_create_table(entity))

    @staticmethod
    def _fqtn(entity: Type[ffd.Entity]):
        return inflection.tableize(entity.get_fqn()).replace('.', '_')

    def _get(self, entity: ffd.Entity):
        if entity.__class__ in self._cache and entity.id_value() in self._cache[entity.__class__]:
            return self._cache[entity.__class__][entity.id_value()]

    def _set(self, entity: ffd.Entity):
        if entity.__class__ not in self._cache:
            self._cache[entity.__class__] = {}
        self._cache[entity.__class__][entity.id_value()] = entity

    def _remove_from_cache(self, entity: ffd.Entity):
        if entity.__class__ in self._cache and entity.id_value() in self._cache[entity.__class__]:
            del self._cache[entity.__class__][entity.id_value()]

    @staticmethod
    def _datetime_declaration(name: str):
        return f"`{name}` timestamp"

    def _generate_update_list(self, entity: Type[ffd.Entity]):
        values = ['obj=:obj']
        for index in self.get_indexes(entity):
            values.append(f'`{index.name}`=:{index.name}')
        return ','.join(values)

    def _build_entity(self, entity: Type[ffd.Entity], data, raw: bool = False):
        return entity.from_dict(self._serializer.deserialize(data[0]['stringValue']))

    def _generate_select_list(self, entity: Type[ffd.Entity]):
        return 'obj'

    def _generate_column_list(self, entity: Type[ffd.Entity]):
        values = ['id', 'obj']
        for index in self.get_indexes(entity):
            values.append(index.name)
        return ','.join(values)

    def _generate_value_list(self, entity: Type[ffd.Entity]):
        placeholders = [':id', ':obj']
        for index in self.get_indexes(entity):
            placeholders.append(f':{index.name}')
        return ','.join(placeholders)

    def _generate_parameters(self, entity: ffd.Entity, part: str = None):
        obj = self._serializer.serialize(entity)
        params = {'id': entity.id_value(), 'obj': obj}
        for field_ in self.get_indexes(entity.__class__):
            params[field_.name] = getattr(entity, field_.name)
        return params

    def _generate_create_table(self, entity: Type[ffd.Entity]):
        columns = []
        indexes = []
        for i in self.get_indexes(entity):
            indexes.append(self._generate_index(i.name))
            if i.type == 'float':
                columns.append(f"`{i.name}` float")
            elif i.type == 'int':
                columns.append(f"`{i.name}` integer")
            elif i.type == 'datetime':
                columns.append(self._datetime_declaration(i.name))
            else:
                length = i.metadata['length'] if 'length' in i.metadata else 256
                columns.append(f"`{i.name}` varchar({length})")
        extra = ''
        if len(columns) > 0:
            self._generate_extra(columns, indexes)
            extra = self._generate_extra(columns, indexes)

        sql = f"""
            create table if not exists {self._fqtn(entity)} (
                id varchar(40)
                , obj longtext not null
                {extra}
                , primary key(id)
            )
        """
        return sql

    def raw(self, entity: Type[ffd.Entity], criteria: ffd.BinaryOp = None, limit: int = None):
        pass



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import inspect
import keyword
import typing
from datetime import datetime

import firefly.domain as ffd

# __pragma__('skip')
from dataclasses import is_dataclass, fields
from abc import ABC
# __pragma__('noskip')
# __pragma__ ('ecom')
from dateparser import parse

"""?
from firefly.presentation.web.polyfills import is_dataclass, fields
?"""
# __pragma__ ('noecom')


keyword_list = keyword.kwlist
keyword_list.append('id')


def _fix_keywords(params: dict):
    if not isinstance(params, dict):
        return params

    for k, v in params.copy().items():
        if k in keyword_list:
            params[f'{k}_'] = v
    return params


def build_argument_list(params: dict, obj: typing.Union[typing.Callable, type]):
    args = {}
    field_dict = {}
    is_dc = False
    params = _fix_keywords(params)

    if is_dataclass(obj):
        is_dc = True
        field_dict = {}
        # noinspection PyDataclass
        for field_ in fields(obj):
            field_dict[field_.name] = field_
        sig = inspect.signature(obj.__init__)
        types = typing.get_type_hints(obj)
    elif isinstance(obj, ffd.MetaAware):
        sig = inspect.signature(obj.__call__)
        types = typing.get_type_hints(obj.__call__)
    elif isinstance(obj, type):
        sig = inspect.signature(obj.__init__)
        types = typing.get_type_hints(obj.__init__)
    else:
        sig = inspect.signature(obj)
        try:
            types = typing.get_type_hints(obj)
        except NameError:
            types = obj.__annotations__

    for param in sig.parameters.values():
        if param.kind == inspect.Parameter.VAR_KEYWORD:
            return params

    for name, param in sig.parameters.items():
        if name == 'self' or param.kind in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD):
            continue

        required = False
        if is_dc:
            required = field_dict[name].metadata.get('required', False) is True
            try:
                d = field_dict[name].default_factory()
                if not isinstance(d, ffd.Empty):
                    required = False
            except (AttributeError, TypeError):
                pass
        elif param.default is not None:
            required = True

        type_ = types[name] if name in types else None
        if type_ is datetime and name in params and isinstance(params[name], str):
            params[name] = parse(params[name]).replace(tzinfo=None)

        if isinstance(type_, type) and issubclass(type_, ffd.ValueObject):
            if name in params and isinstance(params[name], type_):
                args[name] = params[name]
                continue

            try:
                nested = False
                if name in params and isinstance(params[name], dict):
                    entity_args = build_argument_list(params[name], type_)
                    nested = True
                else:
                    entity_args = build_argument_list(params, type_)
            except ffd.MissingArgument:
                if required is False:
                    continue
                raise
            # TODO use factories where appropriate
            args[name] = type_(**entity_args)
            if nested:
                del params[name]
            else:
                for key in entity_args.keys():
                    if not hasattr(type_, 'id_name') or key != type_.id_name():
                        del params[key]
                        if key in args:
                            del args[key]
        elif isinstance(type_, type(typing.List)) and issubclass(type_.__args__[0], ffd.ValueObject):
            args[name] = []
            if isinstance(params, dict) and name in params:
                for d in params[name]:
                    try:
                        entity_args = build_argument_list(d, type_.__args__[0])
                    except ffd.MissingArgument:
                        if required is False:
                            continue
                        raise
                    args[name].append(type_.__args__[0](**entity_args))
        elif isinstance(type_, type(typing.Dict)) and len(type_.__args__) == 2 and \
                issubclass(type_.__args__[1], ffd.ValueObject):
            args[name] = {}
            if isinstance(params, dict) and name in params:
                for k, d in params[name].items():
                    try:
                        entity_args = build_argument_list(d, type_.__args__[1])
                    except ffd.MissingArgument:
                        if required is False:
                            continue
                        raise
                    args[name][k] = type_.__args__[1](**entity_args)
        elif isinstance(params, dict) and name in params:
            args[name] = params[name]
        elif name.endswith('_') and name.rstrip('_') in params:
            args[name] = params[name.rstrip('_')]
        elif required is True:
            raise ffd.MissingArgument(f'Argument: {name} is required')

    return args



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

# __pragma__('skip')
from abc import ABCMeta
from dataclasses import dataclass, fields

import firefly as ff
# __pragma__('noskip')
# __pragma__ ('ecom')
"""?
from firefly.presentation.web.polyfills import ABCMeta, ABC, dataclass, fields
?"""
# __pragma__ ('noecom')

import firefly.domain as ffd


class EntityMeta(ABCMeta):
    def __new__(mcs, name, bases, dct, **kwargs):
        # __pragma__('skip')
        ret = super().__new__(mcs, name, bases, dct)
        # __pragma__('noskip')
        # __pragma__ ('ecom')
        """?
        ret = type.__new__(type, name, bases, dct)
        ?"""
        # __pragma__ ('noecom')
        ret = dataclass(ret, eq=False)

        # __pragma__('skip')
        is_aggregate = False
        for c in bases:
            # We can't import AggregateRoot for a proper issubclass check without getting a circular reference.
            if 'AggregateRoot' in str(c):
                is_aggregate = True
                break

        if is_aggregate:
            if 'create_on' in kwargs:
                ret._create_on = kwargs['create_on']
            if 'delete_on' in kwargs:
                ret._delete_on = kwargs['delete_on']
            if 'update_on' in kwargs:
                ret._update_on = kwargs['update_on']
            if 'sync_with' in kwargs:
                ret._create_on = f"{kwargs['sync_with']}Created"
                ret._delete_on = f"{kwargs['sync_with']}Deleted"
                ret._update_on = f"{kwargs['sync_with']}Updated"
        # __pragma__('noskip')

        return ret

    def __init__(cls, name, bases, dct, **kwargs):
        super().__init__(name, bases, dct)
        cls._c = None

    # __pragma__('skip')
    # noinspection PyDataclass
    @property
    def c(cls):
        if cls._c is None:
            from firefly.domain.repository.search_criteria import AttrFactory
            cls._c = AttrFactory([x.name for x in fields(cls)])
        return cls._c
    # __pragma__('noskip')



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import firefly.domain as ffd
import inflection


@ffd.on(ffd.DeploymentCreated)
class ConfigureDeployment(ffd.ApplicationService):
    _context_map: ffd.ContextMap = None

    def __call__(self, deployment: ffd.Deployment):
        for context in self._context_map.contexts:
            if context.config.get('is_extension', False):
                continue
            service = ffd.Service(name=context.name)
            self._add_gateways(service, context)
            self._add_message_queues(service, context)
            deployment.services.append(service)

    @staticmethod
    def _add_gateways(service: ffd.Service, context: ffd.Context):
        gateway = ffd.ApiGateway()
        name = inflection.dasherize(context.name)
        gateway.endpoints.append(ffd.HttpEndpoint(
            route=f'/{name}',
            method='POST'
        ))
        gateway.endpoints.append(ffd.HttpEndpoint(
            route=f'/{name}',
            method='GET'
        ))

        for endpoint in context.endpoints:
            if isinstance(endpoint, ffd.HttpEndpoint):
                gateway.endpoints.append(endpoint)

        service.api_gateways.append(gateway)

    def _add_message_queues(self, service: ffd.Service, context: ffd.Context):
        top = ffd.NetworkTopology()
        camel = inflection.camelize(context.name)
        subscribers = self._get_subscribers(context)
        top.topics.append(ffd.Topic(name=f'{camel}Topic', subscribers=subscribers))

        service.network_topology = top

    def _get_subscribers(self, context: ffd.Context):
        ret = {}
        for ctx in self._context_map.contexts:
            if ctx.name == context:
                ret[ctx.name] = ctx.queue
                continue

            for service, event_types in ctx.event_listeners.items():
                for event_type in event_types:
                    if isinstance(event_type, str):
                        context_name, _ = event_type.split('.')
                    else:
                        context_name = event_type.get_class_context()
                    if context_name == context.name:
                        ret[context.name] = context.queue

            for service, command_type in ctx.command_handlers.items():
                if isinstance(command_type, str):
                    context_name, _ = command_type.split('.')
                else:
                    context_name = command_type.get_class_context()
                if context_name == context.name:
                    ret[context.name] = context.queue

        return list(ret.values())



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Callable

import firefly.domain as ffd

from .registry import Registry
from ..service.logging.logger import LoggerAware
from ..service.messaging.middleware import Middleware


class TransactionHandlingMiddleware(Middleware, LoggerAware):
    _registry: Registry = None

    def __init__(self):
        self._level = 0

    def __call__(self, message: ffd.Message, next_: Callable) -> ffd.Message:
        try:
            if self._level == 0:
                self.debug('Level 0 - Resetting repositories')
                self.debug(message)
                self._reset()
            self._level += 1
            self.debug('Level incremented: %d', self._level)
            ret = next_(message)
            self._level -= 1
            self.debug('Level decremented: %d', self._level)
            if self._level == 0:
                self.debug('Level 0 - Committing changes')
                self._commit()
            return ret
        except Exception as e:
            self.debug('Caught an exception during transaction')
            self.info(e)
            self._level -= 1
            self.debug('Level decremented: %d', self._level)
            if self._level == 0:
                self.debug('Level 0 - Resetting repositories')
                self._reset()
            raise e

    def _reset(self):
        for repository in self._registry.get_repositories():
            repository.reset()

    def _commit(self):
        for repository in self._registry.get_repositories():
            self.debug('Committing repository %s', repository)
            repository.commit()



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import List, Union, Callable

# __pragma__('skip')
from abc import ABC
# __pragma__('noskip')
# __pragma__ ('ecom')
"""?
from firefly.presentation.web.polyfills import ABC
?"""
# __pragma__ ('noecom')
# __pragma__('opov')

from firefly.domain.service.messaging.middleware_stack import MiddlewareStack
import firefly.domain as ffd


class MessageBus:
    _middleware: list = []
    _handle: MiddlewareStack = None

    def __init__(self, middleware: List[Union[ffd.Middleware, Callable]]):
        self._middleware = middleware
        self._handle = MiddlewareStack(self._middleware)

    @property
    def middleware(self):
        return self._handle.middleware

    @middleware.setter
    def middleware(self, value):
        self._handle.middleware = value

    def add(self, item: Union[ffd.Middleware, Callable]):
        self._handle.add(item)
        
    def insert(self, index: int, item: Union[ffd.Middleware, Callable]):
        self._handle.insert(index, item)

    def replace(self, which: type, with_: ffd.Middleware):
        return self._handle.replace(which, with_)

    def dispatch(self, message: ffd.Message):
        return self._handle(message)


class MessageBusAware(ABC):
    _bus: MessageBus = None

    def dispatch(self, message: ffd.Message):
        return self._bus.dispatch(message)



#!/usr/bin/env python

#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from setuptools import setup

setup(
  name='gyp',
  version='0.1',
  description='Generate Your Projects',
  author='Chromium Authors',
  author_email='chromium-dev@googlegroups.com',
  url='http://code.google.com/p/gyp',
  package_dir = {'': 'pylib'},
  packages=['gyp', 'gyp.generator'],
  entry_points = {'console_scripts': ['gyp=gyp:script_main'] }
)
